#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe guilds on Discord."""

from __future__ import annotations

__all__ = [
    "ActivityAssets",
    "ActivityFlag",
    "ActivitySecret",
    "ActivityTimestamps",
    "ActivityType",
    "ActivityParty",
    "ClientStatus",
    "Guild",
    "GuildEmbed",
    "GuildRole",
    "GuildFeature",
    "GuildSystemChannelFlag",
    "GuildMessageNotificationsLevel",
    "GuildExplicitContentFilterLevel",
    "GuildMFALevel",
    "GuildVerificationLevel",
    "GuildPremiumTier",
    "GuildPreview",
    "GuildMember",
    "GuildMemberPresence",
    "GuildIntegration",
    "GuildMemberBan",
    "IntegrationAccount",
    "IntegrationExpireBehaviour",
    "PartialGuild",
    "PartialGuildIntegration",
    "PartialGuildRole",
    "PresenceActivity",
    "PresenceStatus",
    "PresenceUser",
    "UnavailableGuild",
]

import datetime
import typing

import attr

from hikari import bases
from hikari import channels as _channels
from hikari import colors
from hikari import emojis as _emojis
from hikari import permissions as _permissions
from hikari import unset
from hikari import users
from hikari.internal import conversions
from hikari.internal import marshaller
from hikari.internal import more_enums
from hikari.internal import urls

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


@more_enums.must_be_unique
class GuildExplicitContentFilterLevel(int, more_enums.Enum):
    """Represents the explicit content filter setting for a guild."""

    DISABLED = 0
    """No explicit content filter."""

    MEMBERS_WITHOUT_ROLES = 1
    """Filter posts from anyone without a role."""

    ALL_MEMBERS = 2
    """Filter all posts."""


@more_enums.must_be_unique
class GuildFeature(str, more_enums.Enum):
    """Features that a guild can provide."""

    ANIMATED_ICON = "ANIMATED_ICON"
    """Guild has access to set an animated guild icon."""

    BANNER = "BANNER"
    """Guild has access to set a guild banner image."""

    COMMERCE = "COMMERCE"
    """Guild has access to use commerce features (i.e. create store channels)."""

    DISCOVERABLE = "DISCOVERABLE"
    """Guild is able to be discovered in the directory."""

    FEATURABLE = "FEATURABLE"
    """Guild is able to be featured in the directory."""

    INVITE_SPLASH = "INVITE_SPLASH"
    """Guild has access to set an invite splash background."""

    MORE_EMOJI = "MORE_EMOJI"
    """More emojis can be hosted in this guild than normal."""

    NEWS = "NEWS"
    """Guild has access to create news channels."""

    LURKABLE = "LURKABLE"
    """People can view channels in this guild without joining."""

    PARTNERED = "PARTNERED"
    """Guild is partnered."""

    PUBLIC = "PUBLIC"
    """Guild is public, go figure."""

    PUBLIC_DISABLED = "PUBLIC_DISABLED"
    """Guild cannot be public. Who would have guessed?"""

    VANITY_URL = "VANITY_URL"
    """Guild has access to set a vanity URL."""

    VERIFIED = "VERIFIED"
    """Guild is verified."""

    VIP_REGIONS = "VIP_REGIONS"
    """Guild has access to set 384kbps bitrate in voice.

    Previously gave access to VIP voice servers.
    """

    WELCOME_SCREEN_ENABLED = "WELCOME_SCREEN_ENABLED"
    """Guild has enabled the welcome screen."""


@more_enums.must_be_unique
class GuildMessageNotificationsLevel(int, more_enums.Enum):
    """Represents the default notification level for new messages in a guild."""

    ALL_MESSAGES = 0
    """Notify users when any message is sent."""

    ONLY_MENTIONS = 1
    """Only notify users when they are @mentioned."""


@more_enums.must_be_unique
class GuildMFALevel(int, more_enums.Enum):
    """Represents the multi-factor authorization requirement for a guild."""

    NONE = 0
    """No MFA requirement."""

    ELEVATED = 1
    """MFA requirement."""


@more_enums.must_be_unique
class GuildPremiumTier(int, more_enums.Enum):
    """Tier for Discord Nitro boosting in a guild."""

    NONE = 0
    """No Nitro boost level."""

    TIER_1 = 1
    """Level 1 Nitro boost."""

    TIER_2 = 2
    """Level 2 Nitro boost."""

    TIER_3 = 3
    """Level 3 Nitro boost."""


@more_enums.must_be_unique
class GuildSystemChannelFlag(more_enums.IntFlag):
    """Defines which features are suppressed in the system channel."""

    SUPPRESS_USER_JOIN = 1 << 0
    """Display a message about new users joining."""

    SUPPRESS_PREMIUM_SUBSCRIPTION = 1 << 1
    """Display a message when the guild is Nitro boosted."""


@more_enums.must_be_unique
class GuildVerificationLevel(int, more_enums.Enum):
    """Represents the level of verification of a guild."""

    NONE = 0
    """Unrestricted"""

    LOW = 1
    """Must have a verified email on their account."""

    MEDIUM = 2
    """Must have been registered on Discord for more than 5 minutes."""

    HIGH = 3
    """(╯°□°）╯︵ ┻━┻ - must be a member of the guild for longer than 10 minutes."""

    VERY_HIGH = 4
    """┻━┻ミヽ(ಠ益ಠ)ﾉ彡┻━┻ - must have a verified phone number."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildEmbed(bases.HikariEntity, marshaller.Deserializable):
    """Represents a guild embed."""

    channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, serializer=str, if_none=None, repr=True
    )
    """The ID of the channel the invite for this embed targets, if enabled."""

    is_enabled: bool = marshaller.attrib(raw_name="enabled", deserializer=bool, serializer=bool, repr=True)
    """Whether this embed is enabled."""


def _deserialize_role_ids(payload: more_typing.JSONArray) -> typing.Sequence[bases.Snowflake]:
    return [bases.Snowflake(role_id) for role_id in payload]


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMember(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent a guild bound member."""

    # TODO: make GuildMember delegate to user and implement a common base class
    # this allows members and users to be used interchangeably.
    user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None, inherit_kwargs=True, repr=True
    )
    """This member's user object.

    This will be `None` when attached to Message Create and Update gateway events.
    """

    nickname: typing.Optional[str] = marshaller.attrib(
        raw_name="nick", deserializer=str, if_none=None, if_undefined=None, default=None, repr=True
    )
    """This member's nickname, if set."""

    role_ids: typing.Sequence[bases.Snowflake] = marshaller.attrib(
        raw_name="roles", deserializer=_deserialize_role_ids,
    )
    """A sequence of the IDs of the member's current roles."""

    joined_at: datetime.datetime = marshaller.attrib(deserializer=conversions.parse_iso_8601_ts)
    """The datetime of when this member joined the guild they belong to."""

    premium_since: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_none=None, if_undefined=None, default=None
    )
    """The datetime of when this member started "boosting" this guild.

    This will be `None` if they aren't boosting.
    """

    is_deaf: bool = marshaller.attrib(raw_name="deaf", deserializer=bool)
    """Whether this member is deafened by this guild in it's voice channels."""

    is_mute: bool = marshaller.attrib(raw_name="mute", deserializer=bool)
    """Whether this member is muted by this guild in it's voice channels."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PartialGuildRole(bases.UniqueEntity, marshaller.Deserializable):
    """Represents a partial guild bound Role object."""

    name: str = marshaller.attrib(deserializer=str, serializer=str, repr=True)
    """The role's name."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildRole(PartialGuildRole, marshaller.Serializable):
    """Represents a guild bound Role object."""

    color: colors.Color = marshaller.attrib(
        deserializer=colors.Color, serializer=int, default=colors.Color(0), repr=True,
    )
    """The colour of this role.

    This will be applied to a member's name in chat if it's their top coloured role."""

    is_hoisted: bool = marshaller.attrib(raw_name="hoist", deserializer=bool, serializer=bool, default=False, repr=True)
    """Whether this role is hoisting the members it's attached to in the member list.

    members will be hoisted under their highest role where this is set to `True`."""

    position: int = marshaller.attrib(deserializer=int, serializer=int, default=None, repr=True)
    """The position of this role in the role hierarchy."""

    permissions: _permissions.Permission = marshaller.attrib(
        deserializer=_permissions.Permission, serializer=int, default=_permissions.Permission(0)
    )
    """The guild wide permissions this role gives to the members it's attached to,

    This may be overridden by channel overwrites.
    """

    is_managed: bool = marshaller.attrib(raw_name="managed", deserializer=bool, serializer=None, default=None)
    """Whether this role is managed by an integration."""

    is_mentionable: bool = marshaller.attrib(raw_name="mentionable", deserializer=bool, serializer=bool, default=False)
    """Whether this role can be mentioned by all regardless of permissions."""


@more_enums.must_be_unique
class ActivityType(int, more_enums.Enum):
    """The activity type."""

    PLAYING = 0
    """Shows up as `Playing <name>`"""

    STREAMING = 1

    LISTENING = 2
    """Shows up as `Listening to <name>`."""

    WATCHING = 3
    """Shows up as `Watching <name>`.

    !!! note
        this is not officially documented, so will be likely removed in the near
        future.
    """

    CUSTOM = 4
    """A custom status.

    To set an emoji with the status, place a unicode emoji or Discord emoji
    (`:smiley:`) as the first part of the status activity name.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ActivityTimestamps(bases.HikariEntity, marshaller.Deserializable):
    """The datetimes for the start and/or end of an activity session."""

    start: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.unix_epoch_to_datetime, if_undefined=None, default=None, repr=True
    )
    """When this activity's session was started, if applicable."""

    end: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.unix_epoch_to_datetime, if_undefined=None, default=None, repr=True
    )
    """When this activity's session will end, if applicable."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ActivityParty(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent activity groups of users."""

    id: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None, repr=True)
    """The string id of this party instance, if set."""

    _size_information: typing.Optional[typing.Tuple[int, int]] = marshaller.attrib(
        raw_name="size", deserializer=tuple, if_undefined=None, default=None
    )
    """The size metadata of this party, if applicable."""

    # Ignore docstring not starting in an imperative mood
    @property
    def current_size(self) -> typing.Optional[int]:  # noqa: D401
        """Current size of this party, if applicable."""
        return self._size_information[0] if self._size_information else None

    @property
    def max_size(self) -> typing.Optional[int]:
        """Maximum size of this party, if applicable."""
        return self._size_information[1] if self._size_information else None


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ActivityAssets(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent possible assets for an activity."""

    large_image: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The ID of the asset's large image, if set."""

    large_text: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The text that'll appear when hovering over the large image, if set."""

    small_image: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The ID of the asset's small image, if set."""

    small_text: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The text that'll appear when hovering over the small image, if set."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ActivitySecret(bases.HikariEntity, marshaller.Deserializable):
    """The secrets used for interacting with an activity party."""

    join: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The secret used for joining a party, if applicable."""

    spectate: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The secret used for spectating a party, if applicable."""

    match: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None)
    """The secret used for joining a party, if applicable."""


@more_enums.must_be_unique
class ActivityFlag(more_enums.IntFlag):
    """Flags that describe what an activity includes.

    This can be more than one using bitwise-combinations.
    """

    INSTANCE = 1 << 0
    """Instance"""

    JOIN = 1 << 1
    """Join"""

    SPECTATE = 1 << 2
    """Spectate"""

    JOIN_REQUEST = 1 << 3
    """Join Request"""

    SYNC = 1 << 4
    """Sync"""

    PLAY = 1 << 5
    """Play"""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PresenceActivity(bases.HikariEntity, marshaller.Deserializable):
    """Represents an activity that will be attached to a member's presence."""

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The activity's name."""

    type: ActivityType = marshaller.attrib(deserializer=ActivityType, repr=True)
    """The activity's type."""

    url: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None, default=None)
    """The URL for a `STREAM` type activity, if applicable."""

    created_at: datetime.datetime = marshaller.attrib(deserializer=conversions.unix_epoch_to_datetime)
    """When this activity was added to the user's session."""

    timestamps: typing.Optional[ActivityTimestamps] = marshaller.attrib(
        deserializer=ActivityTimestamps.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The timestamps for when this activity's current state will start and
    end, if applicable.
    """

    application_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None
    )
    """The ID of the application this activity is for, if applicable."""

    details: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None, default=None)
    """The text that describes what the activity's target is doing, if set."""

    state: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None, default=None)
    """The current status of this activity's target, if set."""

    emoji: typing.Union[None, _emojis.UnicodeEmoji, _emojis.UnknownEmoji] = marshaller.attrib(
        deserializer=_emojis.deserialize_reaction_emoji, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The emoji of this activity, if it is a custom status and set."""

    party: typing.Optional[ActivityParty] = marshaller.attrib(
        deserializer=ActivityParty.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """Information about the party associated with this activity, if set."""

    assets: typing.Optional[ActivityAssets] = marshaller.attrib(
        deserializer=ActivityAssets.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """Images and their hover over text for the activity."""

    secrets: typing.Optional[ActivitySecret] = marshaller.attrib(
        deserializer=ActivitySecret.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """Secrets for Rich Presence joining and spectating."""

    is_instance: typing.Optional[bool] = marshaller.attrib(
        raw_name="instance", deserializer=bool, if_undefined=None, default=None
    )
    """Whether this activity is an instanced game session."""

    flags: ActivityFlag = marshaller.attrib(deserializer=ActivityFlag, if_undefined=None, default=None)
    """Flags that describe what the activity includes."""


class PresenceStatus(str, more_enums.Enum):
    """The status of a member."""

    ONLINE = "online"
    """Online/green."""

    IDLE = "idle"
    """Idle/yellow."""

    DND = "dnd"
    """Do not disturb/red."""

    DO_NOT_DISTURB = DND
    """An alias for `PresenceStatus.DND`"""

    OFFLINE = "offline"
    """Offline or invisible/grey."""


def _default_status() -> typing.Literal[PresenceStatus.OFFLINE]:
    return PresenceStatus.OFFLINE


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ClientStatus(bases.HikariEntity, marshaller.Deserializable):
    """The client statuses for this member."""

    desktop: PresenceStatus = marshaller.attrib(
        deserializer=PresenceStatus, if_undefined=_default_status, default=PresenceStatus.OFFLINE, repr=True
    )
    """The status of the target user's desktop session."""

    mobile: PresenceStatus = marshaller.attrib(
        deserializer=PresenceStatus, if_undefined=_default_status, default=PresenceStatus.OFFLINE, repr=True
    )
    """The status of the target user's mobile session."""

    web: PresenceStatus = marshaller.attrib(
        deserializer=PresenceStatus, if_undefined=_default_status, default=PresenceStatus.OFFLINE, repr=True
    )
    """The status of the target user's web session."""


# TODO: should this be an event instead?
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PresenceUser(users.User):
    """A user representation specifically used for presence updates.

    !!! warning
        Every attribute except `PresenceUser.id` may be as `hikari.unset.UNSET`
        unless it is specifically being modified for this update.
    """

    discriminator: typing.Union[str, unset.Unset] = marshaller.attrib(
        deserializer=str, if_undefined=unset.Unset, default=unset.UNSET, repr=True
    )
    """This user's discriminator."""

    username: typing.Union[str, unset.Unset] = marshaller.attrib(
        deserializer=str, if_undefined=unset.Unset, default=unset.UNSET, repr=True
    )
    """This user's username."""

    avatar_hash: typing.Union[None, str, unset.Unset] = marshaller.attrib(
        raw_name="avatar", deserializer=str, if_none=None, if_undefined=unset.Unset, default=unset.UNSET, repr=True
    )
    """This user's avatar hash, if set."""

    is_bot: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="bot", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET, repr=True
    )
    """Whether this user is a bot account."""

    is_system: typing.Union[bool, unset.Unset] = marshaller.attrib(
        raw_name="system", deserializer=bool, if_undefined=unset.Unset, default=unset.UNSET
    )
    """Whether this user is a system account."""

    flags: typing.Union[users.UserFlag, unset.Unset] = marshaller.attrib(
        raw_name="public_flags", deserializer=users.UserFlag, if_undefined=unset.Unset
    )
    """The public flags for this user."""

    @property
    def avatar_url(self) -> typing.Union[str, unset.Unset]:
        """URL for this user's avatar if the relevant info is available.

        !!! note
            This will be `hikari.unset.UNSET` if both `PresenceUser.avatar_hash`
            and `PresenceUser.discriminator` are `hikari.unset.UNSET`.
        """
        return self.format_avatar_url()

    def format_avatar_url(self, fmt: typing.Optional[str] = None, size: int = 4096) -> typing.Union[str, unset.Unset]:
        """Generate the avatar URL for this user's avatar if available.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png` or `gif`.
            Supports `png`, `jpeg`, `jpg`, `webp` and `gif` (when animated).
            Will be ignored for default avatars which can only be `png`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.
            Will be ignored for default avatars.

        Returns
        -------
        typing.Union[str, hikari.unset.UNSET]
            The string URL of the user's custom avatar if
            either `PresenceUser.avatar_hash` is set or their default avatar if
            `PresenceUser.discriminator` is set, else `hikari.unset.UNSET`.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.discriminator is unset.UNSET and self.avatar_hash is unset.UNSET:
            return unset.UNSET
        return super().format_avatar_url(fmt=fmt, size=size)

    @property
    def default_avatar(self) -> typing.Union[int, unset.Unset]:
        """Integer representation of this user's default avatar.

        !!! note
            This will be `hikari.unset.UNSET` if `PresenceUser.discriminator` is
            `hikari.unset.UNSET`.
        """
        if self.discriminator is not unset.UNSET:
            return int(self.discriminator) % 5
        return unset.UNSET


def _deserialize_activities(payload: more_typing.JSONArray, **kwargs: typing.Any) -> typing.Sequence[PresenceActivity]:
    return [PresenceActivity.deserialize(activity, **kwargs) for activity in payload]


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberPresence(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent a guild member's presence."""

    user: PresenceUser = marshaller.attrib(deserializer=PresenceUser.deserialize, inherit_kwargs=True, repr=True)
    """The object of the user who this presence is for.

    !!! info
        Only `PresenceUser.id` is guaranteed for this partial object,
        with other attributes only being included when when they are being
        changed in an event.
    """

    role_ids: typing.Optional[typing.Sequence[bases.Snowflake]] = marshaller.attrib(
        raw_name="roles", deserializer=_deserialize_role_ids, if_undefined=None, default=None,
    )
    """The ids of the user's current roles in the guild this presence belongs to.

    !!! info
        If this is `None` then this information wasn't provided and is unknown.
    """

    guild_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild this presence belongs to.

    This will be `None` when received in an array of members attached to a guild
    object (e.g on Guild Create).
    """

    visible_status: PresenceStatus = marshaller.attrib(raw_name="status", deserializer=PresenceStatus, repr=True)
    """This user's current status being displayed by the client."""

    activities: typing.Sequence[PresenceActivity] = marshaller.attrib(
        deserializer=_deserialize_activities, inherit_kwargs=True
    )
    """An array of the user's activities, with the top one will being
    prioritised by the client.
    """

    client_status: ClientStatus = marshaller.attrib(deserializer=ClientStatus.deserialize, inherit_kwargs=True)
    """An object of the target user's client statuses."""

    premium_since: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=None, if_none=None, default=None
    )
    """The datetime of when this member started "boosting" this guild.

    This will be `None` if they aren't boosting.
    """

    nick: typing.Optional[str] = marshaller.attrib(
        raw_name="nick", deserializer=str, if_undefined=None, if_none=None, default=None, repr=True
    )
    """This member's nickname, if set."""


@more_enums.must_be_unique
class IntegrationExpireBehaviour(int, more_enums.Enum):
    """Behavior for expiring integration subscribers."""

    REMOVE_ROLE = 0
    """Remove the role."""

    KICK = 1
    """Kick the subscriber."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class IntegrationAccount(bases.HikariEntity, marshaller.Deserializable):
    """An account that's linked to an integration."""

    id: str = marshaller.attrib(deserializer=str)
    """The string ID of this (likely) third party account."""

    name: str = marshaller.attrib(deserializer=str)
    """The name of this account."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PartialGuildIntegration(bases.UniqueEntity, marshaller.Deserializable):
    """A partial representation of an integration, found in audit logs."""

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The name of this integration."""

    type: str = marshaller.attrib(deserializer=str, repr=True)
    """The type of this integration."""

    account: IntegrationAccount = marshaller.attrib(deserializer=IntegrationAccount.deserialize, inherit_kwargs=True)
    """The account connected to this integration."""


def _deserialize_expire_grace_period(payload: int) -> datetime.timedelta:
    return datetime.timedelta(days=payload)


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildIntegration(bases.UniqueEntity, marshaller.Deserializable):
    """Represents a guild integration object."""

    is_enabled: bool = marshaller.attrib(raw_name="enabled", deserializer=bool, repr=True)
    """Whether this integration is enabled."""

    is_syncing: bool = marshaller.attrib(raw_name="syncing", deserializer=bool)
    """Whether this integration is syncing subscribers/emojis."""

    role_id: typing.Optional[bases.Snowflake] = marshaller.attrib(deserializer=bases.Snowflake)
    """The ID of the managed role used for this integration's subscribers."""

    is_emojis_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="enable_emoticons", deserializer=bool, if_undefined=None, default=None
    )
    """Whether users under this integration are allowed to use it's custom emojis."""

    expire_behavior: IntegrationExpireBehaviour = marshaller.attrib(deserializer=IntegrationExpireBehaviour)
    """How members should be treated after their connected subscription expires.

    This won't be enacted until after `GuildIntegration.expire_grace_period`
    passes.
    """

    expire_grace_period: datetime.timedelta = marshaller.attrib(deserializer=_deserialize_expire_grace_period,)
    """How many days users with expired subscriptions are given until
    `GuildIntegration.expire_behavior` is enacted out on them
    """

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True)
    """The user this integration belongs to."""

    last_synced_at: datetime.datetime = marshaller.attrib(
        raw_name="synced_at", deserializer=conversions.parse_iso_8601_ts, if_none=None
    )
    """The datetime of when this integration's subscribers were last synced."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildMemberBan(bases.HikariEntity, marshaller.Deserializable):
    """Used to represent guild bans."""

    reason: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None, repr=True)
    """The reason for this ban, will be `None` if no reason was given."""

    user: users.User = marshaller.attrib(deserializer=users.User.deserialize, inherit_kwargs=True, repr=True)
    """The object of the user this ban targets."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UnavailableGuild(bases.UniqueEntity, marshaller.Deserializable):
    """An unavailable guild object, received during gateway events such as READY.

    An unavailable guild cannot be interacted with, and most information may
    be outdated if that is the case.
    """

    # Ignore docstring not starting in an imperative mood
    @property
    def is_unavailable(self) -> bool:  # noqa: D401
        """`True` if this guild is unavailable, else `False`.

        This value is always `True`, and is only provided for consistency.
        """
        return True


def _deserialize_features(payload: more_typing.JSONArray) -> typing.Set[typing.Union[GuildFeature, str]]:
    return {conversions.try_cast(feature, GuildFeature, feature) for feature in payload}


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class PartialGuild(bases.UniqueEntity, marshaller.Deserializable):
    """Base object for any partial guild objects."""

    name: str = marshaller.attrib(deserializer=str, repr=True)
    """The name of the guild."""

    icon_hash: typing.Optional[str] = marshaller.attrib(raw_name="icon", deserializer=str, if_none=None)
    """The hash for the guild icon, if there is one."""

    features: typing.Set[typing.Union[GuildFeature, str]] = marshaller.attrib(deserializer=_deserialize_features)
    """A set of the features in this guild."""

    def format_icon_url(self, fmt: typing.Optional[str] = None, size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's custom icon, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png` or `gif`.
            Supports `png`, `jpeg`, `jpg`, `webp` and `gif` (when
            animated).
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.icon_hash:
            if fmt is None:
                fmt = "gif" if self.icon_hash.startswith("a_") else "png"
            return urls.generate_cdn_url("icons", str(self.id), self.icon_hash, fmt=fmt, size=size)
        return None

    @property
    def icon_url(self) -> typing.Optional[str]:
        """URL for this guild's icon, if set."""
        return self.format_icon_url()


def _deserialize_emojis(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, _emojis.GuildEmoji]:
    return {bases.Snowflake(emoji["id"]): _emojis.GuildEmoji.deserialize(emoji, **kwargs) for emoji in payload}


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class GuildPreview(PartialGuild):
    """A preview of a guild with the `GuildFeature.PUBLIC` feature."""

    splash_hash: typing.Optional[str] = marshaller.attrib(raw_name="splash", deserializer=str, if_none=None)
    """The hash of the splash for the guild, if there is one."""

    discovery_splash_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="discovery_splash", deserializer=str, if_none=None
    )
    """The hash of the discovery splash for the guild, if there is one."""

    emojis: typing.Mapping[bases.Snowflake, _emojis.GuildEmoji] = marshaller.attrib(
        deserializer=_deserialize_emojis, inherit_kwargs=True
    )
    """The mapping of IDs to the emojis this guild provides."""

    approximate_presence_count: int = marshaller.attrib(deserializer=int, repr=True)
    """The approximate amount of presences in guild."""

    approximate_member_count: int = marshaller.attrib(deserializer=int, repr=True)
    """The approximate amount of members in this guild."""

    description: typing.Optional[str] = marshaller.attrib(if_none=None, deserializer=str)
    """The guild's description, if set."""

    def format_splash_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's splash image, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.splash_hash:
            return urls.generate_cdn_url("splashes", str(self.id), self.splash_hash, fmt=fmt, size=size)
        return None

    @property
    def splash_url(self) -> typing.Optional[str]:
        """URL for this guild's splash, if set."""
        return self.format_splash_url()

    def format_discovery_splash_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's discovery splash image, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.discovery_splash_hash:
            return urls.generate_cdn_url(
                "discovery-splashes", str(self.id), self.discovery_splash_hash, fmt=fmt, size=size
            )
        return None

    @property
    def discovery_splash_url(self) -> typing.Optional[str]:
        """URL for this guild's discovery splash, if set."""
        return self.format_discovery_splash_url()


def _deserialize_afk_timeout(payload: int) -> datetime.timedelta:
    return datetime.timedelta(seconds=payload)


def _deserialize_roles(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, GuildRole]:
    return {bases.Snowflake(role["id"]): GuildRole.deserialize(role, **kwargs) for role in payload}


def _deserialize_members(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, GuildMember]:
    return {bases.Snowflake(member["user"]["id"]): GuildMember.deserialize(member, **kwargs) for member in payload}


def _deserialize_channels(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, _channels.GuildChannel]:
    return {bases.Snowflake(channel["id"]): _channels.deserialize_channel(channel, **kwargs) for channel in payload}


def _deserialize_presences(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[bases.Snowflake, GuildMemberPresence]:
    return {
        bases.Snowflake(presence["user"]["id"]): GuildMemberPresence.deserialize(presence, **kwargs)
        for presence in payload
    }


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Guild(PartialGuild):
    """A representation of a guild on Discord.

    !!! note
        If a guild object is considered to be unavailable, then the state of any
        other fields other than the `Guild.is_unavailable` and `Guild.id` are
        outdated or incorrect. If a guild is unavailable, then the contents of
        any other fields should be ignored.
    """

    splash_hash: typing.Optional[str] = marshaller.attrib(raw_name="splash", deserializer=str, if_none=None)
    """The hash of the splash for the guild, if there is one."""

    discovery_splash_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="discovery_splash", deserializer=str, if_none=None
    )
    """The hash of the discovery splash for the guild, if there is one."""

    owner_id: bases.Snowflake = marshaller.attrib(deserializer=bases.Snowflake, repr=True)
    """The ID of the owner of this guild."""

    my_permissions: _permissions.Permission = marshaller.attrib(
        raw_name="permissions", deserializer=_permissions.Permission, if_undefined=None, default=None
    )
    """The guild-level permissions that apply to the bot user.

    This will not take into account permission overwrites or implied
    permissions (for example, ADMINISTRATOR implies all other permissions).

    This will be `None` when this object is retrieved through a REST request
    rather than from the gateway.
    """

    region: str = marshaller.attrib(deserializer=str)
    """The voice region for the guild."""

    afk_channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(deserializer=bases.Snowflake, if_none=None)
    """The ID for the channel that AFK voice users get sent to.

    If `None`, then no AFK channel is set up for this guild.
    """

    afk_timeout: datetime.timedelta = marshaller.attrib(deserializer=_deserialize_afk_timeout)
    """Timeout for activity before a member is classed as AFK.

    How long a voice user has to be AFK for before they are classed as being
    AFK and are moved to the AFK channel (`Guild.afk_channel_id`).
    """

    is_embed_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="embed_enabled", deserializer=bool, if_undefined=False, default=False
    )
    """Defines if the guild embed is enabled or not.

    This information may not be present, in which case, it will be `None`
    instead. This will be `None` for guilds that the bot is not a member in.
    """

    embed_channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, if_none=None, default=None
    )
    """The channel ID that the guild embed will generate an invite to.

    Will be `None` if invites are disabled for this guild's embed.
    """

    verification_level: GuildVerificationLevel = marshaller.attrib(deserializer=GuildVerificationLevel)
    """The verification level required for a user to participate in this guild."""

    default_message_notifications: GuildMessageNotificationsLevel = marshaller.attrib(
        deserializer=GuildMessageNotificationsLevel
    )
    """The default setting for message notifications in this guild."""

    explicit_content_filter: GuildExplicitContentFilterLevel = marshaller.attrib(
        deserializer=GuildExplicitContentFilterLevel
    )
    """The setting for the explicit content filter in this guild."""

    roles: typing.Mapping[bases.Snowflake, GuildRole] = marshaller.attrib(
        deserializer=_deserialize_roles, inherit_kwargs=True
    )
    """The roles in this guild, represented as a mapping of ID to role object."""

    emojis: typing.Mapping[bases.Snowflake, _emojis.GuildEmoji] = marshaller.attrib(
        deserializer=_deserialize_emojis, inherit_kwargs=True
    )
    """A mapping of IDs to the objects of the emojis this guild provides."""

    mfa_level: GuildMFALevel = marshaller.attrib(deserializer=GuildMFALevel)
    """The required MFA level for users wishing to participate in this guild."""

    application_id: typing.Optional[bases.Snowflake] = marshaller.attrib(deserializer=bases.Snowflake, if_none=None)
    """The ID of the application that created this guild.

    This will always be `None` for guilds that weren't created by a bot.
    """

    is_unavailable: typing.Optional[bool] = marshaller.attrib(
        raw_name="unavailable", deserializer=bool, if_undefined=None, default=None
    )
    """Whether the guild is unavailable or not.

    This information is only available if the guild was sent via a
    `GUILD_CREATE` event. If the guild is received from any other place, this
    will always be `None`.

    An unavailable guild cannot be interacted with, and most information may
    be outdated if that is the case.
    """

    is_widget_enabled: typing.Optional[bool] = marshaller.attrib(
        raw_name="widget_enabled", deserializer=bool, if_undefined=None, default=None
    )
    """Describes whether the guild widget is enabled or not.

    If this information is not present, this will be `None`.

    This will only be provided for guilds that the application user is a member
    of. For all other purposes, this should be ignored.
    """

    widget_channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        deserializer=bases.Snowflake, if_undefined=None, if_none=None, default=None
    )
    """The channel ID that the widget's generated invite will send the user to.

    If this information is unavailable or this isn't enabled for the guild then
    this will be `None`.
    """

    system_channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(if_none=None, deserializer=bases.Snowflake)
    """The ID of the system channel or `None` if it is not enabled.

    Welcome messages and Nitro boost messages may be sent to this channel.
    """

    system_channel_flags: GuildSystemChannelFlag = marshaller.attrib(deserializer=GuildSystemChannelFlag)
    """Flags for the guild system channel to describe which notifications are suppressed."""

    rules_channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(if_none=None, deserializer=bases.Snowflake)
    """The ID of the channel where guilds with the `GuildFeature.PUBLIC`
    `features` display rules and guidelines.

    If the `GuildFeature.PUBLIC` feature is not defined, then this is `None`.
    """

    joined_at: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=None, default=None
    )
    """The date and time that the bot user joined this guild.

    This information is only available if the guild was sent via a `GUILD_CREATE`
    event. If the guild is received from any other place, this will always be
    `None`.
    """

    is_large: typing.Optional[bool] = marshaller.attrib(
        raw_name="large", deserializer=bool, if_undefined=None, default=None
    )
    """Whether the guild is considered to be large or not.

    This information is only available if the guild was sent via a `GUILD_CREATE`
    event. If the guild is received from any other place, this will always b
    `None`.

    The implications of a large guild are that presence information will not be
    sent about members who are offline or invisible.
    """

    member_count: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The number of members in this guild.

    This information is only available if the guild was sent via a `GUILD_CREATE`
    event. If the guild is received from any other place, this will always be
    `None`.
    """

    members: typing.Optional[typing.Mapping[bases.Snowflake, GuildMember]] = marshaller.attrib(
        deserializer=_deserialize_members, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """A mapping of ID to the corresponding guild members in this guild.

    This information is only available if the guild was sent via a `GUILD_CREATE`
    event. If the guild is received from any other place, this will always be
    `None`.

    Additionally, any offline members may not be included here, especially if
    there are more members than the large threshold set for the gateway this
    object was send with.

    This information will only be updated if your shards have the correct
    intents set for any update events.

    Essentially, you should not trust the information here to be a full
    representation. If you need complete accurate information, you should
    query the members using the appropriate API call instead.
    """

    channels: typing.Optional[typing.Mapping[bases.Snowflake, _channels.GuildChannel]] = marshaller.attrib(
        deserializer=_deserialize_channels, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """A mapping of ID to the corresponding guild channels in this guild.

    This information is only available if the guild was sent via a `GUILD_CREATE`
    event. If the guild is received from any other place, this will always be
    `None`.

    Additionally, any channels that you lack permissions to see will not be
    defined here.

    This information will only be updated if your shards have the correct
    intents set for any update events.

    To retrieve a list of channels in any other case, you should make an
    appropriate API call to retrieve this information.
    """

    presences: typing.Optional[typing.Mapping[bases.Snowflake, GuildMemberPresence]] = marshaller.attrib(
        deserializer=_deserialize_presences, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """A mapping of member ID to the corresponding presence information for
    the given member, if available.

    This information is only available if the guild was sent via a `GUILD_CREATE`
    event. If the guild is received from any other place, this will always be
    `None`.

    Additionally, any channels that you lack permissions to see will not be
    defined here.

    This information will only be updated if your shards have the correct
    intents set for any update events.

    To retrieve a list of presences in any other case, you should make an
    appropriate API call to retrieve this information.
    """

    max_presences: typing.Optional[int] = marshaller.attrib(
        deserializer=int, if_undefined=None, if_none=None, default=None
    )
    """The maximum number of presences for the guild.

    If this is `None`, then the default value is used (currently 25000).
    """

    max_members: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The maximum number of members allowed in this guild.

    This information may not be present, in which case, it will be `None`.
    """

    max_video_channel_users: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The maximum number of users allowed in a video channel together.

    If not available, this field will be `None`.
    """

    vanity_url_code: typing.Optional[str] = marshaller.attrib(deserializer=str, if_none=None)
    """The vanity URL code for the guild's vanity URL.

    This is only present if `GuildFeature.VANITY_URL` is in `Guild.features` for
    this guild. If not, this will always be `None`.
    """

    description: typing.Optional[str] = marshaller.attrib(if_none=None, deserializer=str)
    """The guild's description.

    This is only present if certain `GuildFeature`'s are set in
    `Guild.features` for this guild. Otherwise, this will always be `None`.
    """

    banner_hash: typing.Optional[str] = marshaller.attrib(raw_name="banner", if_none=None, deserializer=str)
    """The hash for the guild's banner.

    This is only present if the guild has `GuildFeature.BANNER` in
    `Guild.features` for this guild. For all other purposes, it is `None`.
    """

    premium_tier: GuildPremiumTier = marshaller.attrib(deserializer=GuildPremiumTier)
    """The premium tier for this guild."""

    premium_subscription_count: typing.Optional[int] = marshaller.attrib(
        deserializer=int, if_undefined=None, if_none=None, default=None
    )
    """The number of nitro boosts that the server currently has.

    This information may not be present, in which case, it will be `None`.
    """

    preferred_locale: str = marshaller.attrib(deserializer=str)
    """The preferred locale to use for this guild.

    This can only be change if `GuildFeature.PUBLIC` is in `Guild.features`
    for this guild and will otherwise default to `en-US`.
    """

    public_updates_channel_id: typing.Optional[bases.Snowflake] = marshaller.attrib(
        if_none=None, deserializer=bases.Snowflake
    )
    """The channel ID of the channel where admins and moderators receive notices
    from Discord.

    This is only present if `GuildFeature.PUBLIC` is in `Guild.features` for
    this guild. For all other purposes, it should be considered to be `None`.
    """

    # TODO: if this is `None`, then should we attempt to look at the known member count if present?
    approximate_member_count: typing.Optional[int] = marshaller.attrib(
        if_undefined=None, deserializer=int, default=None
    )
    """The approximate number of members in the guild.

    This information will be provided by REST API calls fetching the guilds that
    a bot account is in. For all other purposes, this should be expected to
    remain `None`.
    """

    approximate_active_member_count: typing.Optional[int] = marshaller.attrib(
        raw_name="approximate_presence_count", if_undefined=None, deserializer=int, default=None
    )
    """The approximate number of members in the guild that are not offline.

    This information will be provided by REST API calls fetching the guilds that
    a bot account is in. For all other purposes, this should be expected to
    remain `None`.
    """

    def format_splash_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's splash image, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.splash_hash:
            return urls.generate_cdn_url("splashes", str(self.id), self.splash_hash, fmt=fmt, size=size)
        return None

    @property
    def splash_url(self) -> typing.Optional[str]:
        """URL for this guild's splash, if set."""
        return self.format_splash_url()

    def format_discovery_splash_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's discovery splash image, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.discovery_splash_hash:
            return urls.generate_cdn_url(
                "discovery-splashes", str(self.id), self.discovery_splash_hash, fmt=fmt, size=size
            )
        return None

    @property
    def discovery_splash_url(self) -> typing.Optional[str]:
        """URL for this guild's discovery splash, if set."""
        return self.format_discovery_splash_url()

    def format_banner_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's banner image, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.banner_hash:
            return urls.generate_cdn_url("banners", str(self.id), self.banner_hash, fmt=fmt, size=size)
        return None

    @property
    def banner_url(self) -> typing.Optional[str]:
        """URL for this guild's banner, if set."""
        return self.format_banner_url()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Discord gateway other events."""

from __future__ import annotations

__all__ = [
    "ExceptionEvent",
    "ConnectedEvent",
    "DisconnectedEvent",
    "StartingEvent",
    "StartedEvent",
    "StoppingEvent",
    "StoppedEvent",
    "ReadyEvent",
    "ResumedEvent",
    "UserUpdateEvent",
]

import typing

import attr

from hikari import bases as base_entities
from hikari import guilds
from hikari import users
from hikari.events import base as base_events
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    from hikari.clients import shards  # pylint: disable=cyclic-import
    from hikari.internal import more_typing


# Synthetic event, is not deserialized, and is produced by the dispatcher.
@base_events.no_catch()
@attr.s(slots=True, auto_attribs=True)
class ExceptionEvent(base_events.HikariEvent):
    """Descriptor for an exception thrown while processing an event."""

    exception: Exception
    """The exception that was raised."""

    event: base_events.HikariEvent
    """The event that was being invoked when the exception occurred."""

    callback: typing.Callable[[base_events.HikariEvent], typing.Awaitable[None]]
    """The event that was being invoked when the exception occurred."""


# Synthetic event, is not deserialized
@attr.s(slots=True, auto_attribs=True)
class StartingEvent(base_events.HikariEvent):
    """Event that is fired before the gateway client starts all shards."""


# Synthetic event, is not deserialized
@attr.s(slots=True, auto_attribs=True)
class StartedEvent(base_events.HikariEvent):
    """Event that is fired when the gateway client starts all shards."""


# Synthetic event, is not deserialized
@attr.s(slots=True, auto_attribs=True)
class StoppingEvent(base_events.HikariEvent):
    """Event that is fired when the gateway client is instructed to disconnect all shards."""


# Synthetic event, is not deserialized
@attr.s(slots=True, auto_attribs=True)
class StoppedEvent(base_events.HikariEvent):
    """Event that is fired when the gateway client has finished disconnecting all shards."""


@attr.s(slots=True, kw_only=True, auto_attribs=True)
class ConnectedEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Event invoked each time a shard connects."""

    shard: shards.ShardClient
    """The shard that connected."""


@attr.s(slots=True, kw_only=True, auto_attribs=True)
class DisconnectedEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Event invoked each time a shard disconnects."""

    shard: shards.ShardClient
    """The shard that disconnected."""


@attr.s(slots=True, kw_only=True, auto_attribs=True)
class ResumedEvent(base_events.HikariEvent):
    """Represents a gateway Resume event."""

    shard: shards.ShardClient
    """The shard that reconnected."""


def _deserialize_unavailable_guilds(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[base_entities.Snowflake, guilds.UnavailableGuild]:
    return {
        base_entities.Snowflake(guild["id"]): guilds.UnavailableGuild.deserialize(guild, **kwargs) for guild in payload
    }


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ReadyEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Represents the gateway Ready event.

    This is received only when IDENTIFYing with the gateway.
    """

    gateway_version: int = marshaller.attrib(raw_name="v", deserializer=int, repr=True)
    """The gateway version this is currently connected to."""

    my_user: users.MyUser = marshaller.attrib(
        raw_name="user", deserializer=users.MyUser.deserialize, inherit_kwargs=True, repr=True
    )
    """The object of the current bot account this connection is for."""

    unavailable_guilds: typing.Mapping[base_entities.Snowflake, guilds.UnavailableGuild] = marshaller.attrib(
        raw_name="guilds", deserializer=_deserialize_unavailable_guilds, inherit_kwargs=True
    )
    """A mapping of the guilds this bot is currently in.

    All guilds will start off "unavailable".
    """

    session_id: str = marshaller.attrib(deserializer=str, repr=True)
    """The id of the current gateway session, used for reconnecting."""

    _shard_information: typing.Optional[typing.Tuple[int, int]] = marshaller.attrib(
        raw_name="shard", deserializer=tuple, if_undefined=None, default=None
    )
    """Information about the current shard, only provided when IDENTIFYing."""

    @property
    def shard_id(self) -> typing.Optional[int]:
        """Zero-indexed ID of the current shard.

        This is only available if this ready event was received while IDENTIFYing.
        """
        return self._shard_information[0] if self._shard_information else None

    @property
    def shard_count(self) -> typing.Optional[int]:
        """Total shard count for this bot.

        This is only available if this ready event was received while IDENTIFYing.
        """
        return self._shard_information[1] if self._shard_information else None


# TODO: rename to MyUserUpdateEvent
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class UserUpdateEvent(base_events.HikariEvent, users.MyUser):
    """Used to represent User Update gateway events.

    Sent when the current user is updated.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""A library wide base for storing client components."""

from __future__ import annotations

__all__ = ["Components"]

import typing

import attr

if typing.TYPE_CHECKING:
    from hikari.clients import configs  # pylint: disable=cyclic-import
    from hikari.clients import rest  # pylint: disable=cyclic-import
    from hikari.clients import shards  # pylint: disable=cyclic-import
    from hikari.state import dispatchers  # pylint: disable=cyclic-import
    from hikari.state import event_managers  # pylint: disable=cyclic-import


@attr.s(repr=False)
class Components:
    """A base that defines placement for set of components used in the library."""

    config: configs.BotConfig = attr.attrib(default=None)
    """The config used for this bot."""

    event_dispatcher: dispatchers.EventDispatcher = attr.attrib(default=None)
    """The event dispatcher for this bot."""

    event_manager: event_managers.EventManager = attr.attrib(default=None)
    """The event manager for this bot."""

    rest: rest.RESTClient = attr.attrib(default=None)
    """The REST HTTP client to use for this bot."""

    shards: typing.Mapping[int, shards.ShardClient] = attr.attrib(default=None)
    """Shards registered to this bot.

    These will be created once the bot has started execution.
    """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""An internal marshalling utility used by internal API components.

!!! warning
    You should not change anything in this file, if you do, you will likely get
    unexpected behaviour elsewhere.
"""

from __future__ import annotations

__all__ = [
    "RAISE",
    "dereference_handle",
    "attrib",
    "marshallable",
    "HIKARI_ENTITY_MARSHALLER",
    "HikariEntityMarshaller",
    "Deserializable",
    "Serializable",
]

import importlib
import typing
import weakref

import attr

from hikari.internal import assertions
from hikari.internal import more_collections

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing

_RAW_NAME_ATTR: typing.Final[str] = __name__ + "_RAW_NAME"
_SERIALIZER_ATTR: typing.Final[str] = __name__ + "_SERIALIZER"
_DESERIALIZER_ATTR: typing.Final[str] = __name__ + "_DESERIALIZER"
_INHERIT_KWARGS: typing.Final[str] = __name__ + "_INHERIT_KWARGS"
_IF_UNDEFINED: typing.Final[str] = __name__ + "IF_UNDEFINED"
_IF_NONE: typing.Final[str] = __name__ + "_IF_NONE"
_MARSHALLER_ATTRIB: typing.Final[str] = __name__ + "_MARSHALLER_ATTRIB"
_PASSED_THROUGH_SINGLETONS: typing.Final[typing.Sequence[bool]] = [False, True, None]
RAISE: typing.Final[typing.Any] = object()
EntityT = typing.TypeVar("EntityT", contravariant=True)
ClsT = typing.Type[EntityT]


def dereference_handle(handle_string: str) -> typing.Any:
    """Parse a given handle string into an object reference.

    Parameters
    ----------
    handle_string : str
        The handle to the object to refer to. This is in the format
        `fully.qualified.module.name#object.attribute`. If no `#` is
        input, then the reference will be made to the module itself.

    Returns
    -------
    typing.Any
        The thing that is referred to from this reference.

    Examples
    --------
    * `"collections#deque"`:

        Refers to `collections.deque`

    * `"asyncio.tasks#Task"`:

        Refers to `asyncio.tasks.Task`

    * `"hikari.net"`:

        Refers to `hikari.net`

    * `"foo.bar#baz.bork.qux"`:

        Would refer to a theoretical `qux` attribute on a `bork`
        attribute on a `baz` object in the `foo.bar` module.
    """
    if "#" not in handle_string:
        module, attribute_names = handle_string, ()
    else:
        module, _, attribute_string = handle_string.partition("#")
        attribute_names = attribute_string.split(".")

    obj = importlib.import_module(module)
    for attr_name in attribute_names:
        obj = getattr(obj, attr_name)

    return weakref.proxy(obj)


def attrib(
    *,
    # Mandatory! We do not want to rely on type annotations alone, as they will
    # break if we use __future__.annotations anywhere. If we relied on the
    # field type, that would work, but attrs doesn't let us supply field.type
    # as an attr.ib() kwargs AND use type hints at the same time, and without
    # type hints, the library loses the ability to be type checked properly
    # anymore, so we have to pass this explicitly regardless.
    deserializer: typing.Optional[typing.Callable[[...], typing.Any], type(RAISE)] = RAISE,
    if_none: typing.Union[typing.Callable[[], typing.Any], None, type(RAISE)] = RAISE,
    if_undefined: typing.Union[typing.Callable[[], typing.Any], None, type(RAISE)] = RAISE,
    raw_name: typing.Optional[str] = None,
    inherit_kwargs: bool = False,
    serializer: typing.Optional[typing.Callable[[typing.Any], typing.Any], type(RAISE)] = RAISE,
    **kwargs,
) -> attr.Attribute:
    """Create an `attr.ib` with marshaller metadata attached.

    Parameters
    ----------
    deserializer : typing.Callable[[...], typing.Any], optional
        The deserializer to use to deserialize raw elements.
        If `None` then this field will never be deserialized from a payload
        and will have to be attached to the object after generation or passed
        through to `deserialize` as a kwarg.
    raw_name : str, optional
        The raw name of the element in its raw serialized form. If not provided,
        then this will use the field's default name later.
    inherit_kwargs : bool
        If `True` then any fields passed to deserialize for the entity this
        attribute is attached to as kwargs will also be passed through to this
        entity's deserializer as kwargs. Defaults to `False`.
    if_none : typing.Union[typing.Callable[[], typing.Any], None]
        Either a default factory function called to get the default for when
        this field is `None` or one of `None`, `False` or `True` to specify that
        this should default to the given singleton. Will raise an exception when
        `None` is received for this field later if this isn't specified.
    if_undefined : typing.Union[typing.Callable[[], typing.Any], None]
        Either a default factory function called to get the default for when
        this field isn't defined or one of `None`, `False` or `True` to specify
        that this should default to the given singleton. Will raise an exception
        when this field is undefined later on if this isn't specified.
    serializer : typing.Callable[[typing.Any], typing.Any], optional
        The serializer to use. If not specified, then serializing the entire
        class that this attribute is in will trigger a `TypeError` later.
        If `None` then the field will not be serialized.
    **kwargs :
        Any kwargs to pass to `attr.ib`.

    Returns
    -------
    typing.Any
        The result of `attr.ib` internally being called with additional metadata.
    """
    # Sphinx decides to be really awkward and inject the wrong default values
    # by default. Not helpful when it documents non-optional shit as defaulting
    # to None. Hack to fix this seems to be to turn on autodoc's
    # typing.TYPE_CHECKING mode, and then if that is enabled, always return
    # some dummy class that has a repr that returns a literal "..." string.
    if typing.TYPE_CHECKING:
        return type("Literal", (), {"__repr__": lambda *_: "..."})()

    metadata = kwargs.pop("metadata", {})
    metadata[_RAW_NAME_ATTR] = raw_name
    metadata[_SERIALIZER_ATTR] = serializer
    metadata[_DESERIALIZER_ATTR] = deserializer
    metadata[_IF_NONE] = if_none
    metadata[_IF_UNDEFINED] = if_undefined
    metadata[_INHERIT_KWARGS] = inherit_kwargs
    metadata[_MARSHALLER_ATTRIB] = True

    # Default to not repr-ing a field.
    kwargs.setdefault("repr", False)

    attribute = attr.ib(**kwargs, metadata=metadata)
    # Fool pylint into thinking this is any type.
    return typing.cast(typing.Any, attribute)


def _not_implemented(op, name):
    def error(*_, **__) -> typing.NoReturn:
        raise NotImplementedError(f"Field {name} does not support operation {op}")

    return error


def _default_validator(value: typing.Any):
    assertions.assert_that(
        value is RAISE or value in _PASSED_THROUGH_SINGLETONS or callable(value),
        message=(
            "Invalid default factory passed for `if_undefined` or `if_none`; "
            f"expected a callable or one of the 'passed through singletons' but got {value}."
        ),
        error_type=RuntimeError,
    )


class _AttributeDescriptor:
    __slots__ = (
        "raw_name",
        "field_name",
        "constructor_name",
        "if_none",
        "if_undefined",
        "is_inheriting_kwargs",
        "deserializer",
        "serializer",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        raw_name: str,
        field_name: str,
        constructor_name: str,
        if_none: typing.Union[typing.Callable[..., typing.Any], None, type(RAISE)],
        if_undefined: typing.Union[typing.Callable[..., typing.Any], None, type(RAISE)],
        is_inheriting_kwargs: bool,
        deserializer: typing.Callable[[...], typing.Any],
        serializer: typing.Callable[[typing.Any], typing.Any],
    ) -> None:
        _default_validator(if_undefined)
        _default_validator(if_none)
        self.raw_name = raw_name
        self.field_name = field_name
        self.constructor_name = constructor_name
        self.if_none = if_none
        self.if_undefined = if_undefined
        self.is_inheriting_kwargs = is_inheriting_kwargs
        self.deserializer = deserializer
        self.serializer = serializer


class _EntityDescriptor:
    __slots__ = ("entity_type", "attribs")

    def __init__(self, entity_type: typing.Type, attribs: typing.Collection[_AttributeDescriptor]) -> None:
        self.entity_type = entity_type
        self.attribs = tuple(attribs)


def _construct_attribute_descriptor(field: attr.Attribute) -> _AttributeDescriptor:
    raw_name = typing.cast(str, field.metadata.get(_RAW_NAME_ATTR) or field.name)
    field_name = typing.cast(str, field.name)

    constructor_name = field_name

    # Attrs strips leading underscores for generated __init__ methods.
    while constructor_name.startswith("_"):
        constructor_name = constructor_name[1:]

    deserializer = field.metadata[_DESERIALIZER_ATTR]
    serializer = field.metadata[_SERIALIZER_ATTR]

    return _AttributeDescriptor(
        raw_name=raw_name,
        field_name=field_name,
        constructor_name=constructor_name,
        if_none=field.metadata[_IF_NONE],
        if_undefined=field.metadata[_IF_UNDEFINED],
        is_inheriting_kwargs=field.metadata[_INHERIT_KWARGS],
        deserializer=deserializer if deserializer is not RAISE else _not_implemented("deserialize", field_name),
        serializer=serializer if serializer is not RAISE else _not_implemented("serialize", field_name),
    )


def _construct_entity_descriptor(entity: typing.Any) -> _EntityDescriptor:
    assertions.assert_that(
        hasattr(entity, "__attrs_attrs__"),
        f"{entity.__module__}.{entity.__qualname__} is not an attr class",
        error_type=TypeError,
    )

    return _EntityDescriptor(
        entity,
        [
            _construct_attribute_descriptor(field)
            for field in attr.fields(entity)
            if field.metadata.get(_MARSHALLER_ATTRIB)
        ],
    )


class HikariEntityMarshaller:
    """Hikari's utility to manage automated serialization and deserialization.

    It can deserialize and serialize any internal components that that are
    decorated with the `marshallable` decorator, and that are
    `attr.s` classes using fields with the`attrib` function call descriptor.
    """

    __slots__ = ("_registered_entities",)

    def __init__(self) -> None:
        self._registered_entities: typing.MutableMapping[typing.Type, _EntityDescriptor] = {}

    def register(self, cls: typing.Type[EntityT]) -> typing.Type[EntityT]:
        """Register an attrs type for fast future deserialization.

        Parameters
        ----------
        cls : typing.Type[typing.Any]
            The type to register.

        Returns
        -------
        typing.Type[typing.Any]
            The input argument. This enables this to be used as a decorator if
            desired.

        Raises
        ------
        TypeError
            If the class is not an `attr.s` class.
        """
        entity_descriptor = _construct_entity_descriptor(cls)
        self._registered_entities[cls] = entity_descriptor
        return cls

    def deserialize(
        self, raw_data: more_typing.JSONObject, target_type: typing.Type[EntityT], **injected_kwargs: typing.Any
    ) -> EntityT:
        """Deserialize a given raw data item into the target type.

        Parameters
        ----------
        raw_data : typing.Mapping[str, typing.Any]
            The raw data to deserialize.
        target_type : typing.Type[typing.Any]
            The type to deserialize to.
        **injected_kwargs :
            Attributes to inject into the entity. These still need to be
            included in the model's slots and should normally be fields where
            both `deserializer` and `serializer` are set to `None`.

        Returns
        -------
        typing.Any
            The deserialized instance.

        Raises
        ------
        LookupError
            If the entity is not registered.
        AttributeError
            If the field is not optional, but the field was not present in the
            raw payload, or it was present, but it was assigned `None`.
        TypeError
            If the deserialization call failed for some reason.
        """
        try:
            descriptor = self._registered_entities[target_type]
        except KeyError:
            raise LookupError(f"No registered entity {target_type.__module__}.{target_type.__qualname__}")

        kwargs = {}

        for a in descriptor.attribs:
            if a.deserializer is None:
                continue
            kwarg_name = a.constructor_name

            if a.raw_name not in raw_data:
                if a.if_undefined is RAISE:
                    raise AttributeError(
                        "Failed to deserialize data to instance of "
                        f"{target_type.__module__}.{target_type.__qualname__} due to required field {a.field_name} "
                        f"(from raw key {a.raw_name}) not being included in the input payload\n\n{raw_data}"
                    )
                if a.if_undefined in _PASSED_THROUGH_SINGLETONS:
                    kwargs[kwarg_name] = a.if_undefined
                else:
                    kwargs[kwarg_name] = a.if_undefined()
                continue

            if (data := raw_data[a.raw_name]) is None:
                if a.if_none is RAISE:
                    raise AttributeError(
                        "Failed to deserialize data to instance of "
                        f"{target_type.__module__}.{target_type.__qualname__} due to non-nullable field {a.field_name}"
                        f" (from raw key {a.raw_name}) being `None` in the input payload\n\n{raw_data}"
                    )
                if a.if_none in _PASSED_THROUGH_SINGLETONS:
                    kwargs[kwarg_name] = a.if_none
                else:
                    kwargs[kwarg_name] = a.if_none()
                continue

            try:
                kwargs[kwarg_name] = a.deserializer(
                    data, **(injected_kwargs if a.is_inheriting_kwargs else more_collections.EMPTY_DICT)
                )
            except Exception as exc:
                raise TypeError(
                    "Failed to deserialize data to instance of "
                    f"{target_type.__module__}.{target_type.__qualname__} because marshalling failed on "
                    f"attribute {a.field_name} (passed to constructor as {kwarg_name})"
                ) from exc

        return target_type(**kwargs, **injected_kwargs)

    def serialize(self, obj: typing.Optional[typing.Any]) -> more_typing.NullableJSONObject:
        """Serialize a given entity into a raw data item.

        Parameters
        ----------
        obj : typing.Any, optional
            The entity to serialize.

        Returns
        -------
        typing.Mapping[str, typing.Any], optional
            The serialized raw data item.

        Raises
        ------
        LookupError
            If the entity is not registered.
        """
        if obj is None:
            return None

        input_type = type(obj)

        try:
            descriptor = self._registered_entities[input_type]
        except KeyError:
            raise LookupError(f"No registered entity {input_type.__module__}.{input_type.__qualname__}")

        raw_data = {}

        for a in descriptor.attribs:
            if a.serializer is None:
                continue
            if (value := getattr(obj, a.field_name)) is not None:
                raw_data[a.raw_name] = a.serializer(value)

        return raw_data


HIKARI_ENTITY_MARSHALLER = HikariEntityMarshaller()


def marshallable(*, marshaller: HikariEntityMarshaller = HIKARI_ENTITY_MARSHALLER) -> typing.Callable[[ClsT], ClsT]:
    """Create a decorator for a class to make it into an `attr.s` class.

    Parameters
    ----------
    marshaller : HikariEntityMarshaller
        If specified, this should be an instance of a marshaller to use. For
        most internal purposes, you want to not specify this, since it will
        then default to the hikari-global marshaller instead. This is
        useful, however, for testing and for external usage.

    Returns
    -------
    `decorator(T) -> T`
        A decorator to decorate a class with.

    !!! note
        The `auto_attribs` functionality provided by `attr.s` is not
        supported by this marshaller utility. Do not attempt to use it!

    Examples
    --------
        @attrs()
        class MyEntity:
            id: int = attrib(deserializer=int, serializer=str)
            password: str = attrib(deserializer=int, transient=True)
            ...
    """

    def decorator(cls: ClsT) -> ClsT:
        marshaller.register(cls)
        return cls

    return decorator


class Deserializable:
    """Mixin that enables the class to be deserialized from a raw entity."""

    __slots__ = ()

    @classmethod
    def deserialize(
        cls: typing.Type[more_typing.T_contra], payload: more_typing.JSONType, **kwargs
    ) -> more_typing.T_contra:
        """Deserialize the given payload into the object.

        Parameters
        ----------
        payload
            The payload to deserialize into the object.
        """
        return HIKARI_ENTITY_MARSHALLER.deserialize(payload, cls, **kwargs)


class Serializable:
    """Mixin that enables an instance of the class to be serialized."""

    __slots__ = ()

    def serialize(self: more_typing.T_contra) -> more_typing.JSONType:
        """Serialize this instance into a naive value."""
        return HIKARI_ENTITY_MARSHALLER.serialize(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Event management for stateless bots."""

from __future__ import annotations

__all__ = ["StatelessEventManagerImpl"]

from hikari.events import channel
from hikari.events import guild
from hikari.events import message
from hikari.events import other
from hikari.state import dispatchers
from hikari.state import event_managers


# pylint: disable=too-many-public-methods
class StatelessEventManagerImpl(event_managers.EventManager[dispatchers.EventDispatcher]):
    """Stateless event manager implementation for stateless bots.

    This is an implementation that does not rely on querying prior information to
    operate. The implementation details of this are much simpler than a stateful
    version, and are not immediately affected by the use of intents.
    """

    @event_managers.raw_event_mapper("CONNECTED")
    def on_connect(self, shard, _) -> None:
        """Handle CONNECTED events."""
        event = other.ConnectedEvent(shard=shard)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("DISCONNECTED")
    def on_disconnect(self, shard, _) -> None:
        """Handle DISCONNECTED events."""
        event = other.DisconnectedEvent(shard=shard)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("RESUME")
    def on_resume(self, shard, _) -> None:
        """Handle RESUME events."""
        event = other.ResumedEvent(shard=shard)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("READY")
    def on_ready(self, _, payload) -> None:
        """Handle READY events."""
        event = other.ReadyEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("CHANNEL_CREATE")
    def on_channel_create(self, _, payload) -> None:
        """Handle CHANNEL_CREATE events."""
        event = channel.ChannelCreateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("CHANNEL_UPDATE")
    def on_channel_update(self, _, payload) -> None:
        """Handle CHANNEL_UPDATE events."""
        event = channel.ChannelUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("CHANNEL_DELETE")
    def on_channel_delete(self, _, payload) -> None:
        """Handle CHANNEL_DELETE events."""
        event = channel.ChannelDeleteEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("CHANNEL_PIN_UPDATE")
    def on_channel_pin_update(self, _, payload) -> None:
        """Handle CHANNEL_PIN_UPDATE events."""
        event = channel.ChannelPinUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_CREATE")
    def on_guild_create(self, _, payload) -> None:
        """Handle GUILD_CREATE events."""
        event = guild.GuildCreateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_UPDATE")
    def on_guild_update(self, _, payload) -> None:
        """Handle GUILD_UPDATE events."""
        event = guild.GuildUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_DELETE")
    def on_guild_delete(self, _, payload) -> None:
        """Handle GUILD_DELETE events."""
        if payload.get("unavailable", False):
            event = guild.GuildUnavailableEvent.deserialize(payload, components=self._components)
        else:
            event = guild.GuildLeaveEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_BAN_ADD")
    def on_guild_ban_add(self, _, payload) -> None:
        """Handle GUILD_BAN_ADD events."""
        event = guild.GuildBanAddEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_BAN_REMOVE")
    def on_guild_ban_remove(self, _, payload) -> None:
        """Handle GUILD_BAN_REMOVE events."""
        event = guild.GuildBanRemoveEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_EMOJIS_UPDATE")
    def on_guild_emojis_update(self, _, payload) -> None:
        """Handle GUILD_EMOJIS_UPDATE events."""
        event = guild.GuildEmojisUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_INTEGRATIONS_UPDATE")
    def on_guild_integrations_update(self, _, payload) -> None:
        """Handle GUILD_INTEGRATIONS_UPDATE events."""
        event = guild.GuildIntegrationsUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_MEMBER_ADD")
    def on_guild_member_add(self, _, payload) -> None:
        """Handle GUILD_MEMBER_ADD events."""
        event = guild.GuildMemberAddEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_MEMBER_UPDATE")
    def on_guild_member_update(self, _, payload) -> None:
        """Handle GUILD_MEMBER_UPDATE events."""
        event = guild.GuildMemberUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_MEMBER_REMOVE")
    def on_guild_member_remove(self, _, payload) -> None:
        """Handle GUILD_MEMBER_REMOVE events."""
        event = guild.GuildMemberRemoveEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_ROLE_CREATE")
    def on_guild_role_create(self, _, payload) -> None:
        """Handle GUILD_ROLE_CREATE events."""
        event = guild.GuildRoleCreateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_ROLE_UPDATE")
    def on_guild_role_update(self, _, payload) -> None:
        """Handle GUILD_ROLE_UPDATE events."""
        event = guild.GuildRoleUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("GUILD_ROLE_DELETE")
    def on_guild_role_delete(self, _, payload) -> None:
        """Handle GUILD_ROLE_DELETE events."""
        event = guild.GuildRoleDeleteEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("INVITE_CREATE")
    def on_invite_create(self, _, payload) -> None:
        """Handle INVITE_CREATE events."""
        event = channel.InviteCreateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("INVITE_DELETE")
    def on_invite_delete(self, _, payload) -> None:
        """Handle INVITE_DELETE events."""
        event = channel.InviteDeleteEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_CREATE")
    def on_message_create(self, _, payload) -> None:
        """Handle MESSAGE_CREATE events."""
        event = message.MessageCreateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_UPDATE")
    def on_message_update(self, _, payload) -> None:
        """Handle MESSAGE_UPDATE events."""
        event = message.MessageUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_DELETE")
    def on_message_delete(self, _, payload) -> None:
        """Handle MESSAGE_DELETE events."""
        event = message.MessageDeleteEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_DELETE_BULK")
    def on_message_delete_bulk(self, _, payload) -> None:
        """Handle MESSAGE_DELETE_BULK events."""
        event = message.MessageDeleteBulkEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_REACTION_ADD")
    def on_message_reaction_add(self, _, payload) -> None:
        """Handle MESSAGE_REACTION_ADD events."""
        event = message.MessageReactionAddEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_REACTION_REMOVE")
    def on_message_reaction_remove(self, _, payload) -> None:
        """Handle MESSAGE_REACTION_REMOVE events."""
        payload["emoji"].setdefault("animated", None)

        event = message.MessageReactionRemoveEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("MESSAGE_REACTION_REMOVE_EMOJI")
    def on_message_reaction_remove_emoji(self, _, payload) -> None:
        """Handle MESSAGE_REACTION_REMOVE_EMOJI events."""
        payload["emoji"].setdefault("animated", None)

        event = message.MessageReactionRemoveEmojiEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("PRESENCE_UPDATE")
    def on_presence_update(self, _, payload) -> None:
        """Handle PRESENCE_UPDATE events."""
        event = guild.PresenceUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("TYPING_START")
    def on_typing_start(self, _, payload) -> None:
        """Handle TYPING_START events."""
        event = channel.TypingStartEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("USER_UPDATE")
    def on_user_update(self, _, payload) -> None:
        """Handle USER_UPDATE events."""
        event = other.UserUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("VOICE_STATE_UPDATE")
    def on_voice_state_update(self, _, payload) -> None:
        """Handle VOICE_STATE_UPDATE events."""
        event = channel.VoiceStateUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("VOICE_SERVER_UPDATE")
    def on_voice_server_update(self, _, payload) -> None:
        """Handle VOICE_SERVER_UPDATE events."""
        event = channel.VoiceStateUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)

    @event_managers.raw_event_mapper("WEBHOOK_UPDATE")
    def on_webhook_update(self, _, payload) -> None:
        """Handle WEBHOOK_UPDATE events."""
        event = channel.WebhookUpdateEvent.deserialize(payload, components=self._components)
        self._components.event_dispatcher.dispatch_event(event)


# pylint: enable=too-many-public-methods



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""The logic for handling all requests to oauth2 endpoints."""

from __future__ import annotations

__all__ = ["RESTOAuth2Component"]

import abc

from hikari import applications
from hikari.clients.rest import base


class RESTOAuth2Component(base.BaseRESTComponent, abc.ABC):  # pylint: disable=abstract-method
    """The REST client component for handling requests to oauth2 endpoints."""

    async def fetch_my_application_info(self) -> applications.Application:
        """Get the current application information.

        Returns
        -------
        hikari.applications.Application
            An application info object.
        """
        payload = await self._session.get_current_application_info()
        return applications.Application.deserialize(payload, components=self._components)

    async def add_guild_member(self, *_, **__):
        # TODO: implement and document this.
        # https://discord.com/developers/docs/resources/guild#add-guild-member
        raise NotImplementedError()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe message embeds on Discord."""

from __future__ import annotations

__all__ = [
    "Embed",
    "EmbedThumbnail",
    "EmbedVideo",
    "EmbedImage",
    "EmbedProvider",
    "EmbedAuthor",
    "EmbedFooter",
    "EmbedField",
]

import datetime
import typing

import attr

from hikari import bases
from hikari import colors
from hikari import files
from hikari.internal import assertions
from hikari.internal import conversions
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing

_MAX_FOOTER_TEXT: typing.Final[int] = 2048
_MAX_AUTHOR_NAME: typing.Final[int] = 256
_MAX_FIELD_NAME: typing.Final[int] = 256
_MAX_FIELD_VALUE: typing.Final[int] = 1024
_MAX_EMBED_TITLE: typing.Final[int] = 256
_MAX_EMBED_DESCRIPTION: typing.Final[int] = 2048
_MAX_EMBED_FIELDS: typing.Final[int] = 25
_MAX_EMBED_SIZE: typing.Final[int] = 6000


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedFooter(bases.HikariEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents an embed footer."""

    text: str = marshaller.attrib(deserializer=str, serializer=str, repr=True)
    """The footer text."""

    icon_url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None
    )
    """The URL of the footer icon."""

    proxy_icon_url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=None, if_undefined=None, default=None
    )
    """The proxied URL of the footer icon.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedImage(bases.HikariEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents an embed image."""

    url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None, repr=True,
    )
    """The URL of the image."""

    proxy_url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=None, if_undefined=None, default=None,
    )
    """The proxied URL of the image.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """

    height: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=None, if_undefined=None, default=None)
    """The height of the image.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """

    width: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=None, if_undefined=None, default=None)
    """The width of the image.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedThumbnail(bases.HikariEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents an embed thumbnail."""

    url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None, repr=True,
    )
    """The URL of the thumbnail."""

    proxy_url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=None, if_undefined=None, default=None,
    )
    """The proxied URL of the thumbnail.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """

    height: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=None, if_undefined=None, default=None)
    """The height of the thumbnail.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """

    width: typing.Optional[int] = marshaller.attrib(deserializer=int, serializer=None, if_undefined=None, default=None)
    """The width of the thumbnail.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedVideo(bases.HikariEntity, marshaller.Deserializable):
    """Represents an embed video.

    !!! note
        This embed attached object cannot be sent by bots or webhooks while
        sending an embed and therefore shouldn't be initiated like the other
        embed objects.
    """

    url: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None, repr=True)
    """The URL of the video."""

    height: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The height of the video."""

    width: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The width of the video."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedProvider(bases.HikariEntity, marshaller.Deserializable):
    """Represents an embed provider.

    !!! note
        This embed attached object cannot be sent by bots or webhooks while
        sending an embed and therefore shouldn't be sent by your application.
        You should still expect to receive these objects where appropriate.
    """

    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None, repr=True)
    """The name of the provider."""

    url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, if_undefined=None, if_none=None, default=None, repr=True
    )
    """The URL of the provider."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedAuthor(bases.HikariEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents an embed author."""

    name: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None, repr=True
    )
    """The name of the author."""

    url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None, repr=True
    )
    """The URL of the author."""

    icon_url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None
    )
    """The URL of the author icon."""

    proxy_icon_url: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=None, if_undefined=None, default=None
    )
    """The proxied URL of the author icon.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class EmbedField(bases.HikariEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents a field in a embed."""

    name: str = marshaller.attrib(deserializer=str, serializer=str, repr=True)
    """The name of the field."""

    value: str = marshaller.attrib(deserializer=str, serializer=str, repr=True)
    """The value of the field."""

    is_inline: bool = marshaller.attrib(
        raw_name="inline", deserializer=bool, serializer=bool, if_undefined=False, default=False, repr=True
    )
    """Whether the field should display inline. Defaults to `False`."""


def _serialize_timestamp(timestamp: datetime.datetime) -> str:
    return timestamp.replace(tzinfo=datetime.timezone.utc).isoformat()


def _deserialize_fields(payload: more_typing.JSONArray, **kwargs: typing.Any) -> typing.Sequence[EmbedField]:
    return [EmbedField.deserialize(field, **kwargs) for field in payload]


def _serialize_fields(fields: typing.Sequence[EmbedField]) -> more_typing.JSONArray:
    return [field.serialize() for field in fields]


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Embed(bases.HikariEntity, marshaller.Deserializable, marshaller.Serializable):
    """Represents an embed."""

    title: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None, repr=True
    )
    """The title of the embed."""

    @title.validator
    def _title_check(self, _, value):  # pylint:disable=unused-argument
        if value is not None:
            assertions.assert_that(
                len(value) <= _MAX_EMBED_TITLE, f"title must not exceed {_MAX_EMBED_TITLE} characters"
            )

    description: typing.Optional[str] = marshaller.attrib(
        deserializer=str, serializer=str, if_undefined=None, default=None
    )
    """The description of the embed."""

    @description.validator
    def _description_check(self, _, value):  # pylint:disable=unused-argument
        if value is not None:
            assertions.assert_that(
                len(value) <= _MAX_EMBED_DESCRIPTION, f"description must not exceed {_MAX_EMBED_DESCRIPTION} characters"
            )

    url: typing.Optional[str] = marshaller.attrib(deserializer=str, serializer=str, if_undefined=None, default=None)
    """The URL of the embed."""

    timestamp: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts,
        serializer=_serialize_timestamp,
        if_undefined=None,
        default=None,
        repr=True,
    )
    """The timestamp of the embed."""

    color: typing.Optional[colors.Color] = marshaller.attrib(
        deserializer=colors.Color,
        serializer=int,
        converter=attr.converters.optional(colors.Color.of),
        if_undefined=None,
        default=None,
    )
    """The colour of this embed's sidebar."""

    footer: typing.Optional[EmbedFooter] = marshaller.attrib(
        deserializer=EmbedFooter.deserialize,
        serializer=EmbedFooter.serialize,
        if_undefined=None,
        default=None,
        inherit_kwargs=True,
    )
    """The footer of the embed."""

    image: typing.Optional[EmbedImage] = marshaller.attrib(
        deserializer=EmbedImage.deserialize,
        serializer=EmbedImage.serialize,
        if_undefined=None,
        default=None,
        inherit_kwargs=True,
    )
    """The image of the embed."""

    thumbnail: typing.Optional[EmbedThumbnail] = marshaller.attrib(
        deserializer=EmbedThumbnail.deserialize,
        serializer=EmbedThumbnail.serialize,
        if_undefined=None,
        default=None,
        inherit_kwargs=True,
    )
    """The thumbnail of the embed."""

    video: typing.Optional[EmbedVideo] = marshaller.attrib(
        deserializer=EmbedVideo.deserialize, serializer=None, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The video of the embed.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """

    provider: typing.Optional[EmbedProvider] = marshaller.attrib(
        deserializer=EmbedProvider.deserialize, serializer=None, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """The provider of the embed.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed and
        will be ignored during serialization.
    """

    author: typing.Optional[EmbedAuthor] = marshaller.attrib(
        deserializer=EmbedAuthor.deserialize,
        serializer=EmbedAuthor.serialize,
        if_undefined=None,
        default=None,
        inherit_kwargs=True,
    )
    """The author of the embed."""

    fields: typing.Sequence[EmbedField] = marshaller.attrib(
        deserializer=_deserialize_fields,
        serializer=_serialize_fields,
        if_undefined=list,
        factory=list,
        inherit_kwargs=True,
    )
    """The fields of the embed."""

    _assets_to_upload = attr.attrib(factory=list)

    @property
    def assets_to_upload(self):
        """File assets that need to be uploaded when sending the embed."""
        return self._assets_to_upload

    @staticmethod
    def _extract_url(url) -> typing.Tuple[typing.Optional[str], typing.Optional[files.BaseStream]]:
        if url is None:
            return None, None
        if isinstance(url, files.BaseStream):
            return f"attachment://{url.filename}", url
        return url, None

    def _maybe_ref_file_obj(self, file_obj) -> None:
        if file_obj is not None:
            self._assets_to_upload.append(file_obj)

    def set_footer(self, *, text: str, icon: typing.Optional[str, files.BaseStream] = None) -> Embed:
        """Set the embed footer.

        Parameters
        ----------
        text: str
            The optional text to set for the footer.
        icon: typing.Union[str, hikari.files.BaseStream], optional
            The optional `hikari.files.BaseStream` or URL to the image to set.

        Returns
        -------
        Embed
            This embed to allow method chaining.

        Raises
        ------
        ValueError
            If `text` exceeds 2048 characters or consists purely of whitespaces.
        """
        assertions.assert_that(len(text.strip()) > 0, "footer.text must not be empty or purely of whitespaces")
        assertions.assert_that(
            len(text) <= _MAX_FOOTER_TEXT, f"footer.text must not exceed {_MAX_FOOTER_TEXT} characters"
        )
        icon, file = self._extract_url(icon)
        self.footer = EmbedFooter(text=text, icon_url=icon)
        self._maybe_ref_file_obj(file)
        return self

    def set_image(self, image: typing.Optional[str, files.BaseStream] = None) -> Embed:
        """Set the embed image.

        Parameters
        ----------
        image: typing.Union[str, hikari.files.BaseStream], optional
            The optional `hikari.files.BaseStream` or URL to the image to set.

        Returns
        -------
        Embed
            This embed to allow method chaining.
        """
        image, file = self._extract_url(image)
        self.image = EmbedImage(url=image)
        self._maybe_ref_file_obj(file)
        return self

    def set_thumbnail(self, image: typing.Optional[str, files.BaseStream] = None) -> Embed:
        """Set the thumbnail image.

        Parameters
        ----------
        image: typing.Union[str, hikari.files.BaseStream], optional
            The optional `hikari.files.BaseStream` or URL to the image to set.

        Returns
        -------
        Embed
            This embed to allow method chaining.
        """
        image, file = self._extract_url(image)
        self.thumbnail = EmbedThumbnail(url=image)
        self._maybe_ref_file_obj(file)
        return self

    def set_author(
        self,
        *,
        name: typing.Optional[str] = None,
        url: typing.Optional[str] = None,
        icon: typing.Optional[str, files.BaseStream] = None,
    ) -> Embed:
        """Set the author of this embed.

        Parameters
        ----------
        name: str, optional
            The optional authors name.
        url: str, optional
            The optional URL to make the author text link to.
        icon: typing.Union[str, hikari.files.BaseStream], optional
            The optional `hikari.files.BaseStream` or URL to the icon to set.

        Returns
        -------
        Embed
            This embed to allow method chaining.

        Raises
        ------
        ValueError
            If `name` exceeds 256 characters or consists purely of whitespaces.
        """
        assertions.assert_that(
            name is None or len(name.strip()) > 0, "author.name must not be empty or purely of whitespaces"
        )
        assertions.assert_that(
            name is None or len(name) <= _MAX_AUTHOR_NAME, f"author.name must not exceed {_MAX_AUTHOR_NAME} characters"
        )
        icon, icon_file = self._extract_url(icon)
        self.author = EmbedAuthor(name=name, url=url, icon_url=icon)
        self._maybe_ref_file_obj(icon_file)
        return self

    def add_field(self, *, name: str, value: str, inline: bool = False, index: typing.Optional[int] = None) -> Embed:
        """Add a field to this embed.

        Parameters
        ----------
        name: str
            The fields name (title).
        value: str
            The fields value.
        inline: bool
            Whether to set the field to behave as if it were inline or not. Defaults to `False`.
        index: int, optional
            The optional index to insert the field at. If `None`, it will append to the end.

        Returns
        -------
        Embed
            This embed to allow method chaining.

        Raises
        ------
        ValueError
            If `title` exceeds 256 characters or `value` exceeds 2048 characters; if
            the `name` or `value` consist purely of whitespace, or be zero characters in size;
            25 fields are present in the embed.
        """
        index = index if index is not None else len(self.fields)
        assertions.assert_that(
            len(self.fields) <= _MAX_EMBED_FIELDS and index < _MAX_EMBED_FIELDS,
            f"no more than {_MAX_EMBED_FIELDS} fields can be stored",
        )
        assertions.assert_that(len(name.strip()) > 0, "field.name must not be empty or purely of whitespaces")
        assertions.assert_that(len(value.strip()) > 0, "field.value must not be empty or purely of whitespaces")
        assertions.assert_that(
            len(name.strip()) <= _MAX_FIELD_NAME, f"field.name must not exceed {_MAX_FIELD_NAME} characters"
        )
        assertions.assert_that(
            len(value) <= _MAX_FIELD_VALUE, f"field.value must not exceed {_MAX_FIELD_VALUE} characters"
        )

        self.fields.insert(index, EmbedField(name=name, value=value, is_inline=inline))
        return self

    def edit_field(self, index: int, *, name: str = ..., value: str = ..., inline: bool = ...) -> Embed:
        """Edit a field in this embed at the given index.

        Parameters
        ----------
        index: int
            The index to edit the field at.
        name: str
            If specified, the new fields name (title).
        value: str
            If specified, the new fields value.
        inline: bool
            If specified, the whether to set the field to behave as if it were
            inline or not.

        Returns
        -------
        Embed
            This embed to allow method chaining.

        Raises
        ------
        IndexError
            If you referred to an index that doesn't exist.
        ValueError
            If `title` exceeds 256 characters or `value` exceeds 2048 characters; if
            the `name` or `value` consist purely of whitespace, or be zero characters in size.
        """
        if name is not ...:
            assertions.assert_that(len(name.strip()) > 0, "field.name must not be empty or purely of whitespaces")
            assertions.assert_that(
                len(name.strip()) <= _MAX_FIELD_NAME, f"field.name must not exceed {_MAX_FIELD_NAME} characters"
            )
        if value is not ...:
            assertions.assert_that(len(value.strip()) > 0, "field.value must not be empty or purely of whitespaces")
            assertions.assert_that(
                len(value) <= _MAX_FIELD_VALUE, f"field.value must not exceed {_MAX_FIELD_VALUE} characters"
            )

        field = self.fields[index]

        field.name = name if name is not ... else field.name
        field.value = value if value is not ... else field.value
        field.is_inline = inline if value is not ... else field.is_inline
        return self

    def remove_field(self, index: int) -> Embed:
        """Remove a field from this embed at the given index.

        Parameters
        ----------
        index: int
            The index of the field to remove.

        Returns
        -------
        Embed
            This embed to allow method chaining.

        Raises
        ------
        IndexError
            If you referred to an index that doesn't exist.
        """
        del self.fields[index]
        return self

    @staticmethod
    def _safe_len(item) -> int:
        return len(item) if item is not None else 0

    def _check_total_length(self) -> None:
        total_size = self._safe_len(self.title)
        total_size += self._safe_len(self.description)
        total_size += self._safe_len(self.author.name) if self.author is not None else 0
        total_size += len(self.footer.text) if self.footer is not None else 0

        for field in self.fields:
            total_size += len(field.name)
            total_size += len(field.value)

        assertions.assert_that(
            total_size <= _MAX_EMBED_SIZE, f"Total characters in an embed can not exceed {_MAX_EMBED_SIZE}"
        )

    def serialize(self) -> more_typing.JSONObject:
        self._check_total_length()
        return super().serialize()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Marshall wrappings for the REST implementation in `hikari.net.rest`.

This provides an object-oriented interface for interacting with discord's REST
API.
"""

from __future__ import annotations

__all__ = ["RESTClient"]

import typing

from hikari.clients.rest import channel
from hikari.clients.rest import gateway
from hikari.clients.rest import guild
from hikari.clients.rest import invite
from hikari.clients.rest import me
from hikari.clients.rest import oauth2
from hikari.clients.rest import react
from hikari.clients.rest import user
from hikari.clients.rest import voice
from hikari.clients.rest import webhook
from hikari.net import rest as low_level_rest

if typing.TYPE_CHECKING:
    from hikari.clients import components as _components


class RESTClient(
    channel.RESTChannelComponent,
    me.RESTCurrentUserComponent,
    gateway.RESTGatewayComponent,
    guild.RESTGuildComponent,
    invite.RESTInviteComponent,
    oauth2.RESTOAuth2Component,
    react.RESTReactionComponent,
    user.RESTUserComponent,
    voice.RESTVoiceComponent,
    webhook.RESTWebhookComponent,
):
    """
    A marshalling object-oriented REST API client.

    This client bridges the basic REST API exposed by
    `hikari.net.rest.REST` and wraps it in a unit of processing that can handle
    handle parsing API objects into Hikari entity objects.

    Parameters
    ----------
    components : hikari.clients.components.Components
        The client components that this rest client should be bound by.
        Includes the rest config.

    !!! note
        For all endpoints where a `reason` argument is provided, this may be a
        string inclusively between `0` and `512` characters length, with any
        additional characters being cut off.
    """

    def __init__(self, components: _components.Components) -> None:
        token = None
        if components.config.token_type is not None:
            token = f"{components.config.token_type} {components.config.token}"
        super().__init__(
            components,
            low_level_rest.REST(
                allow_redirects=components.config.allow_redirects,
                base_url=components.config.rest_url,
                connector=components.config.tcp_connector,
                debug=components.config.debug,
                proxy_headers=components.config.proxy_headers,
                proxy_auth=components.config.proxy_auth,
                ssl_context=components.config.ssl_context,
                verify_ssl=components.config.verify_ssl,
                timeout=components.config.request_timeout,
                token=token,
                trust_env=components.config.trust_env,
                version=components.config.rest_version,
            ),
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe Discord gateway channel events."""
from __future__ import annotations

__all__ = [
    "BaseChannelEvent",
    "ChannelCreateEvent",
    "ChannelUpdateEvent",
    "ChannelDeleteEvent",
    "ChannelPinUpdateEvent",
    "WebhookUpdateEvent",
    "TypingStartEvent",
    "InviteCreateEvent",
    "InviteDeleteEvent",
    "VoiceStateUpdateEvent",
    "VoiceServerUpdateEvent",
]

import abc
import datetime
import typing

import attr

from hikari import bases as base_entities
from hikari import channels
from hikari import guilds
from hikari import intents
from hikari import invites
from hikari import users
from hikari import voices
from hikari.events import base as base_events
from hikari.internal import conversions
from hikari.internal import marshaller

if typing.TYPE_CHECKING:
    from hikari.internal import more_typing


def _overwrite_deserializer(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[base_entities.Snowflake, channels.PermissionOverwrite]:
    return {
        base_entities.Snowflake(overwrite["id"]): channels.PermissionOverwrite.deserialize(overwrite, **kwargs)
        for overwrite in payload
    }


def _rate_limit_per_user_deserializer(seconds: int) -> datetime.timedelta:
    return datetime.timedelta(seconds=seconds)


def _recipients_deserializer(
    payload: more_typing.JSONArray, **kwargs: typing.Any
) -> typing.Mapping[base_entities.Snowflake, users.User]:
    return {base_entities.Snowflake(user["id"]): users.User.deserialize(user, **kwargs) for user in payload}


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class BaseChannelEvent(base_events.HikariEvent, base_entities.UniqueEntity, marshaller.Deserializable, abc.ABC):
    """A base object that Channel events will inherit from."""

    type: channels.ChannelType = marshaller.attrib(deserializer=channels.ChannelType, repr=True)
    """The channel's type."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild this channel is in, will be `None` for DMs."""

    position: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The sorting position of this channel.

    This will be relative to the `BaseChannelEvent.parent_id` if set.
    """

    permission_overwrites: typing.Optional[
        typing.Mapping[base_entities.Snowflake, channels.PermissionOverwrite]
    ] = marshaller.attrib(deserializer=_overwrite_deserializer, if_undefined=None, default=None, inherit_kwargs=True)
    """An mapping of the set permission overwrites for this channel, if applicable."""

    name: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, default=None, repr=True)
    """The name of this channel, if applicable."""

    topic: typing.Optional[str] = marshaller.attrib(deserializer=str, if_undefined=None, if_none=None, default=None)
    """The topic of this channel, if applicable and set."""

    is_nsfw: typing.Optional[bool] = marshaller.attrib(
        raw_name="nsfw", deserializer=bool, if_undefined=None, default=None
    )
    """Whether this channel is nsfw, will be `None` if not applicable."""

    last_message_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_none=None, if_undefined=None, default=None
    )
    """The ID of the last message sent, if it's a text type channel."""

    bitrate: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The bitrate (in bits) of this channel, if it's a guild voice channel."""

    user_limit: typing.Optional[int] = marshaller.attrib(deserializer=int, if_undefined=None, default=None)
    """The user limit for this channel if it's a guild voice channel."""

    rate_limit_per_user: typing.Optional[datetime.timedelta] = marshaller.attrib(
        deserializer=_rate_limit_per_user_deserializer, if_undefined=None, default=None
    )
    """How long a user has to wait before sending another message in this channel.

    This is only applicable to a guild text like channel.
    """

    recipients: typing.Optional[typing.Mapping[base_entities.Snowflake, users.User]] = marshaller.attrib(
        deserializer=_recipients_deserializer, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """A mapping of this channel's recipient users, if it's a DM or group DM."""

    icon_hash: typing.Optional[str] = marshaller.attrib(
        raw_name="icon", deserializer=str, if_undefined=None, if_none=None, default=None
    )
    """The hash of this channel's icon, if it's a group DM channel and is set."""

    owner_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None
    )
    """The ID of this channel's creator, if it's a DM channel."""

    application_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None
    )
    """The ID of the application that created the group DM.

    This is only applicable to bot based group DMs.
    """

    parent_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, if_none=None, default=None
    )
    """The ID of this channels's parent category within guild, if set."""

    last_pin_timestamp: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=None, default=None
    )
    """The datetime of when the last message was pinned in this channel."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelCreateEvent(BaseChannelEvent):
    """Represents Channel Create gateway events.

    Will be sent when a guild channel is created and before all Create Message
    events that originate from a DM channel.
    """


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelUpdateEvent(BaseChannelEvent):
    """Represents Channel Update gateway events."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelDeleteEvent(BaseChannelEvent):
    """Represents Channel Delete gateway events."""


@base_events.requires_intents(intents.Intent.GUILDS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class ChannelPinUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent the Channel Pins Update gateway event.

    Sent when a message is pinned or unpinned in a channel but not
    when a pinned message is deleted.
    """

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild where this event happened.

    Will be `None` if this happened in a DM channel.
    """

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel where the message was pinned or unpinned."""

    last_pin_timestamp: typing.Optional[datetime.datetime] = marshaller.attrib(
        deserializer=conversions.parse_iso_8601_ts, if_undefined=None, default=None, repr=True
    )
    """The datetime of when the most recent message was pinned in this channel.

    Will be `None` if there are no messages pinned after this change.
    """


@base_events.requires_intents(intents.Intent.GUILD_WEBHOOKS)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class WebhookUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent webhook update gateway events.

    Sent when a webhook is updated, created or deleted in a guild.
    """

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild this webhook is being updated in."""

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel this webhook is being updated in."""


def _timestamp_deserializer(date: str) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(date, datetime.timezone.utc)


@base_events.requires_intents(intents.Intent.GUILD_MESSAGE_TYPING, intents.Intent.DIRECT_MESSAGE_TYPING)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class TypingStartEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent typing start gateway events.

    Received when a user or bot starts "typing" in a channel.
    """

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel this typing event is occurring in."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild this typing event is occurring in.

    Will be `None` if this event is happening in a DM channel.
    """

    user_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the user who triggered this typing event."""

    timestamp: datetime.datetime = marshaller.attrib(deserializer=_timestamp_deserializer)
    """The datetime of when this typing event started."""

    member: typing.Optional[guilds.GuildMember] = marshaller.attrib(
        deserializer=guilds.GuildMember.deserialize, if_undefined=None, default=None
    )
    """The member object of the user who triggered this typing event.

    Will be `None` if this was triggered in a DM.
    """


def _max_age_deserializer(age: int) -> typing.Optional[datetime.datetime]:
    return datetime.timedelta(seconds=age) if age > 0 else None


def _max_uses_deserializer(count: int) -> typing.Union[int, float]:
    return count or float("inf")


@base_events.requires_intents(intents.Intent.GUILD_INVITES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class InviteCreateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Represents a gateway Invite Create event."""

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel this invite targets."""

    code: str = marshaller.attrib(deserializer=str, repr=True)
    """The code that identifies this invite."""

    created_at: datetime.datetime = marshaller.attrib(deserializer=conversions.parse_iso_8601_ts)
    """The datetime of when this invite was created."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild this invite was created in, if applicable.

    Will be `None` for group DM invites.
    """

    inviter: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The object of the user who created this invite, if applicable."""

    max_age: typing.Optional[datetime.timedelta] = marshaller.attrib(deserializer=_max_age_deserializer,)
    """The timedelta of how long this invite will be valid for.

    If set to `None` then this is unlimited.
    """

    max_uses: typing.Union[int, float] = marshaller.attrib(deserializer=_max_uses_deserializer)
    """The limit for how many times this invite can be used before it expires.

    If set to infinity (`float("inf")`) then this is unlimited.
    """

    target_user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None, inherit_kwargs=True
    )
    """The object of the user who this invite targets, if set."""

    target_user_type: typing.Optional[invites.TargetUserType] = marshaller.attrib(
        deserializer=invites.TargetUserType, if_undefined=None, default=None
    )
    """The type of user target this invite is, if applicable."""

    is_temporary: bool = marshaller.attrib(raw_name="temporary", deserializer=bool)
    """Whether this invite grants temporary membership."""

    uses: int = marshaller.attrib(deserializer=int)
    """The amount of times this invite has been used."""


@base_events.requires_intents(intents.Intent.GUILD_INVITES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class InviteDeleteEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent Invite Delete gateway events.

    Sent when an invite is deleted for a channel we can access.
    """

    channel_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the channel this ID was attached to."""

    # TODO: move common fields with InviteCreateEvent into base class.
    code: str = marshaller.attrib(deserializer=str, repr=True)
    """The code of this invite."""

    guild_id: typing.Optional[base_entities.Snowflake] = marshaller.attrib(
        deserializer=base_entities.Snowflake, if_undefined=None, default=None, repr=True
    )
    """The ID of the guild this invite was deleted in.

    This will be `None` if this invite belonged to a DM channel.
    """


@base_events.requires_intents(intents.Intent.GUILD_VOICE_STATES)
@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class VoiceStateUpdateEvent(base_events.HikariEvent, voices.VoiceState):
    """Used to represent voice state update gateway events.

    Sent when a user joins, leaves or moves voice channel(s).
    """


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class VoiceServerUpdateEvent(base_events.HikariEvent, marshaller.Deserializable):
    """Used to represent voice server update gateway events.

    Sent when initially connecting to voice and when the current voice instance
    falls over to a new server.
    """

    token: str = marshaller.attrib(deserializer=str)
    """The voice connection's string token."""

    guild_id: base_entities.Snowflake = marshaller.attrib(deserializer=base_entities.Snowflake, repr=True)
    """The ID of the guild this voice server update is for."""

    endpoint: str = marshaller.attrib(deserializer=str, repr=True)
    """The uri for this voice server host."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Enumerations for opcodes and status codes."""

from __future__ import annotations

__all__ = ["GatewayCloseCode", "GatewayOpcode", "JSONErrorCode"]

from hikari.internal import more_enums


@more_enums.must_be_unique
class GatewayCloseCode(int, more_enums.Enum):
    """Reasons for closing a gateway connection.

    !!! note
        Any codes greater than or equal to `4000` are server-side codes. Any
        codes between `1000` and `1999` inclusive are generally client-side codes.
    """

    NORMAL_CLOSURE = 1000
    """The application running closed."""

    UNKNOWN_ERROR = 4000
    """Discord is not sure what went wrong. Try reconnecting?"""

    UNKNOWN_OPCODE = 4001
    """You sent an invalid Gateway opcode or an invalid payload for an opcode.

    Don't do that!
    """

    DECODE_ERROR = 4002
    """You sent an invalid payload to Discord. Don't do that!"""

    NOT_AUTHENTICATED = 4003
    """You sent Discord a payload prior to IDENTIFYing."""

    AUTHENTICATION_FAILED = 4004
    """The account token sent with your identify payload is incorrect."""

    ALREADY_AUTHENTICATED = 4005
    """You sent more than one identify payload. Don't do that!"""

    INVALID_SEQ = 4007
    """The sequence sent when resuming the session was invalid.

    Reconnect and start a new session.
    """

    RATE_LIMITED = 4008
    """Woah nelly! You're sending payloads to Discord too quickly. Slow it down!"""

    SESSION_TIMEOUT = 4009
    """Your session timed out. Reconnect and start a new one."""

    INVALID_SHARD = 4010
    """You sent Discord an invalid shard when IDENTIFYing."""

    SHARDING_REQUIRED = 4011
    """The session would have handled too many guilds.

    You are required to shard your connection in order to connect.
    """

    INVALID_VERSION = 4012
    """You sent an invalid version for the gateway."""

    INVALID_INTENT = 4013
    """You sent an invalid intent for a Gateway Intent.

    You may have incorrectly calculated the bitwise value.
    """

    DISALLOWED_INTENT = 4014
    """You sent a disallowed intent for a Gateway Intent.

    You may have tried to specify an intent that you have not enabled or are not
    whitelisted for.
    """

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title()
        return f"{self.value} {name}"


@more_enums.must_be_unique
class GatewayOpcode(int, more_enums.Enum):
    """Opcodes that the gateway uses internally."""

    DISPATCH = 0
    """An event was dispatched."""

    HEARTBEAT = 1
    """Used for ping checking."""

    IDENTIFY = 2
    """Used for client handshake."""

    PRESENCE_UPDATE = 3
    """Used to update the client status."""

    VOICE_STATE_UPDATE = 4
    """Used to join/move/leave voice channels."""

    RESUME = 6
    """Used to resume a closed connection."""

    RECONNECT = 7
    """Used to tell clients to reconnect to the gateway."""

    REQUEST_GUILD_MEMBERS = 8
    """Used to request guild members."""

    INVALID_SESSION = 9
    """Used to notify client they have an invalid session id."""

    HELLO = 10
    """Sent immediately after connecting.

    Contains heartbeat and server debug information.
    """

    HEARTBEAT_ACK = 11
    """Sent immediately following a client heartbeat that was received."""

    GUILD_SYNC = 12
    """Not yet documented, so do not use."""

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title()
        return f"{self.value} {name}"


@more_enums.must_be_unique
class JSONErrorCode(int, more_enums.Enum):
    """Error codes that can be returned by the REST API."""

    GENERAL_ERROR = 0
    """This is sent if the payload is screwed up, etc."""

    UNKNOWN_ACCOUNT = 10_001
    """Unknown account"""

    UNKNOWN_APPLICATION = 10_002
    """Unknown application"""

    UNKNOWN_CHANNEL = 10_003
    """Unknown channel"""

    UNKNOWN_GUILD = 10_004
    """Unknown guild"""

    UNKNOWN_INTEGRATION = 10_005
    """Unknown integration"""

    UNKNOWN_INVITE = 10_006
    """Unknown invite"""

    UNKNOWN_MEMBER = 10_007
    """Unknown member"""

    UNKNOWN_MESSAGE = 10_008
    """Unknown message"""

    UNKNOWN_OVERWRITE = 10_009
    """Unknown overwrite"""

    UNKNOWN_PROVIDER = 10_010
    """Unknown provider"""

    UNKNOWN_ROLE = 10_011
    """Unknown role"""

    UNKNOWN_TOKEN = 10_012
    """Unknown token"""

    UNKNOWN_USER = 10_013
    """Unknown user"""

    UNKNOWN_EMOJI = 10_014
    """Unknown emoji"""

    UNKNOWN_WEBHOOK = 10_015
    """Unknown Webhook"""

    UNKNOWN_BAN = 10_026
    """Unknown ban"""

    USERS_ONLY = 20_001
    """Bots cannot use this endpoint

    !!! note
        You should never expect to receive this in normal API usage.
    """

    BOTS_ONLY = 20_002
    """Only bots can use this endpoint.

    !!! note
        You should never expect to receive this in normal API usage.
    """

    MAX_GUILDS_REACHED = 30_001
    """Maximum number of guilds reached (100)

    !!! note
        You should never expect to receive this in normal API usage as this only
        applies to user accounts.

    This is unlimited for bot accounts.
    """

    MAX_FRIENDS_REACHED = 30_002
    """Maximum number of friends reached (1000)

    !!! note
        You should never expect to receive this in normal API usage as this only
        applies to user accounts.

        Bots cannot have friends :( .
    """

    MAX_PINS_REACHED = 30_003
    """Maximum number of pins reached (50)"""

    MAX_GUILD_ROLES_REACHED = 30_005
    """Maximum number of guild roles reached (250)"""

    MAX_WEBHOOKS_REACHED = 30_007
    """Maximum number of webhooks reached (10)"""

    MAX_REACTIONS_REACHED = 30_010
    """Maximum number of reactions reached (20)"""

    MAX_GUILD_CHANNELS_REACHED = 30_013
    """Maximum number of guild channels reached (500)"""

    MAX_MESSAGE_ATTACHMENTS_REACHED = 30_015
    """Maximum number of attachments in a message reached (10)"""

    MAX_INVITES_REACHED = 30_016
    """Maximum number of invites reached (10000)"""

    NEEDS_VERIFICATION = 40_002
    """You need to verify your account to perform this action."""

    UNAUTHORIZED = 40_001
    """Unauthorized"""

    TOO_LARGE = 40_005
    """Request entity too large. Try sending something smaller in size"""

    DISABLED_TEMPORARILY = 40_006
    """This feature has been temporarily disabled server-side"""

    USER_BANNED = 40_007
    """The user is banned from this guild"""

    MISSING_ACCESS = 50_001
    """Missing access"""

    INVALID_ACCOUNT_TYPE = 50_002
    """Invalid account type"""

    CANNOT_EXECUTE_ACTION_ON_DM_CHANNEL = 50_003
    """Cannot execute action on a DM channel"""

    WIDGET_DISABLED = 50_004
    """Widget Disabled"""

    CANNOT_EDIT_A_MESSAGE_AUTHORED_BY_ANOTHER_USER = 50_005
    """Cannot edit a message authored by another user"""

    CANNOT_SEND_AN_EMPTY_MESSAGE = 50_006
    """Cannot send an empty message"""

    CANNOT_SEND_MESSAGES_TO_THIS_USER = 50_007
    """Cannot send messages to this user"""

    CANNOT_SEND_MESSAGES_IN_VOICE_CHANNEL = 50_008
    """Cannot send messages in a voice channel"""

    CHANNEL_VERIFICATION_TOO_HIGH = 50_009
    """Channel verification level is too high"""

    OAUTH2_APPLICATION_DOES_NOT_HAVE_A_BOT = 50_010
    """OAuth2 application does not have a bot"""

    OAUTH2_APPLICATION_LIMIT_REACHED = 50_011
    """OAuth2 application limit reached"""

    INVALID_OAUTH2_STATE = 50_012
    """Invalid OAuth state"""

    MISSING_PERMISSIONS = 50_013
    """Missing permissions"""

    INVALID_AUTHENTICATION_TOKEN = 50_014
    """Invalid authentication token"""

    NOTE_IS_TOO_LONG = 50_015
    """Note is too long"""

    INVALID_NUMBER_OF_MESSAGES_TO_DELETE = 50_016
    """Provided too few or too many messages to delete.

    Must provide at least 2 and fewer than 100 messages to delete.
    """

    CANNOT_PIN_A_MESSAGE_IN_A_DIFFERENT_CHANNEL = 50_019
    """A message can only be pinned to the channel it was sent in"""

    INVALID_INVITE = 50_020
    """Invite code is either invalid or taken."""

    CANNOT_EXECUTE_ACTION_ON_SYSTEM_MESSAGE = 50_021
    """Cannot execute action on a system message"""

    INVALID_OAUTH2_TOKEN = 50_025
    """Invalid OAuth2 access token"""

    MESSAGE_PROVIDED_WAS_TOO_OLD_TO_BULK_DELETE = 50_034
    """A message provided was too old to bulk delete"""

    INVALID_FORM_BODY = 50_035
    """Invalid Form Body"""

    ACCEPTED_INVITE_TO_GUILD_BOT_IS_NOT_IN = 50_036
    """An invite was accepted to a guild the application's bot is not in"""

    INVALID_API_VERSION = 50_041
    """Invalid API version"""

    REACTION_BLOCKED = 90_001
    """Reaction blocked"""

    RESOURCE_OVERLOADED = 130_000
    """The resource is overloaded."""

    def __str__(self) -> str:
        name = self.name.replace("_", " ").title()
        return f"{self.value} {name}"


# pylint: enable=no-member



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekoka.tt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Components and entities that are used to describe invites on Discord."""

from __future__ import annotations

__all__ = ["TargetUserType", "VanityUrl", "InviteGuild", "Invite", "InviteWithMetadata"]

import datetime
import typing

import attr

from hikari import bases
from hikari import channels
from hikari import guilds
from hikari import users
from hikari.internal import conversions
from hikari.internal import marshaller
from hikari.internal import more_enums
from hikari.internal import urls


@more_enums.must_be_unique
class TargetUserType(int, more_enums.Enum):
    """The reason a invite targets a user."""

    STREAM = 1
    """This invite is targeting a "Go Live" stream."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class VanityUrl(bases.HikariEntity, marshaller.Deserializable):
    """A special case invite object, that represents a guild's vanity url."""

    code: str = marshaller.attrib(deserializer=str, repr=True)
    """The code for this invite."""

    uses: int = marshaller.attrib(deserializer=int, repr=True)
    """The amount of times this invite has been used."""


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class InviteGuild(guilds.PartialGuild):
    """Represents the partial data of a guild that'll be attached to invites."""

    splash_hash: typing.Optional[str] = marshaller.attrib(raw_name="splash", deserializer=str, if_none=None)
    """The hash of the splash for the guild, if there is one."""

    banner_hash: typing.Optional[str] = marshaller.attrib(raw_name="banner", if_none=None, deserializer=str)
    """The hash for the guild's banner.

    This is only present if `hikari.guilds.GuildFeature.BANNER` is in the
    `features` for this guild. For all other purposes, it is `None`.
    """

    description: typing.Optional[str] = marshaller.attrib(if_none=None, deserializer=str)
    """The guild's description.

    This is only present if certain `features` are set in this guild.
    Otherwise, this will always be `None`. For all other purposes, it is `None`.
    """

    verification_level: guilds.GuildVerificationLevel = marshaller.attrib(deserializer=guilds.GuildVerificationLevel)
    """The verification level required for a user to participate in this guild."""

    vanity_url_code: typing.Optional[str] = marshaller.attrib(if_none=None, deserializer=str, repr=True)
    """The vanity URL code for the guild's vanity URL.

    This is only present if `hikari.guilds.GuildFeature.VANITY_URL` is in the
    `features` for this guild. If not, this will always be `None`.
    """

    def format_splash_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's splash, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.splash_hash:
            return urls.generate_cdn_url("splashes", str(self.id), self.splash_hash, fmt=fmt, size=size)
        return None

    @property
    def splash_url(self) -> typing.Optional[str]:
        """URL for this guild's splash, if set."""
        return self.format_splash_url()

    def format_banner_url(self, fmt: str = "png", size: int = 4096) -> typing.Optional[str]:
        """Generate the URL for this guild's banner, if set.

        Parameters
        ----------
        fmt : str
            The format to use for this URL, defaults to `png`.
            Supports `png`, `jpeg`, `jpg` and `webp`.
        size : int
            The size to set for the URL, defaults to `4096`.
            Can be any power of two between 16 and 4096.

        Returns
        -------
        str, optional
            The string URL.

        Raises
        ------
        ValueError
            If `size` is not a power of two or not between 16 and 4096.
        """
        if self.banner_hash:
            return urls.generate_cdn_url("banners", str(self.id), self.banner_hash, fmt=fmt, size=size)
        return None

    @property
    def banner_url(self) -> typing.Optional[str]:
        """URL for this guild's banner, if set."""
        return self.format_banner_url()


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class Invite(bases.HikariEntity, marshaller.Deserializable):
    """Represents an invite that's used to add users to a guild or group dm."""

    code: str = marshaller.attrib(deserializer=str, repr=True)
    """The code for this invite."""

    guild: typing.Optional[InviteGuild] = marshaller.attrib(
        deserializer=InviteGuild.deserialize, if_undefined=None, default=None, inherit_kwargs=True, repr=True
    )
    """The partial object of the guild this dm belongs to.

    Will be `None` for group dm invites.
    """

    channel: channels.PartialChannel = marshaller.attrib(
        deserializer=channels.PartialChannel.deserialize, inherit_kwargs=True, repr=True
    )
    """The partial object of the channel this invite targets."""

    inviter: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """The object of the user who created this invite."""

    target_user: typing.Optional[users.User] = marshaller.attrib(
        deserializer=users.User.deserialize, if_undefined=None, default=None, inherit_kwargs=True,
    )
    """The object of the user who this invite targets, if set."""

    target_user_type: typing.Optional[TargetUserType] = marshaller.attrib(
        deserializer=TargetUserType, if_undefined=None, default=None
    )
    """The type of user target this invite is, if applicable."""

    approximate_presence_count: typing.Optional[int] = marshaller.attrib(
        deserializer=int, if_undefined=None, default=None
    )
    """The approximate amount of presences in this invite's guild.

    This is only present when `with_counts` is passed as `True` to the GET
    Invites endpoint.
    """

    approximate_member_count: typing.Optional[int] = marshaller.attrib(
        deserializer=int, if_undefined=None, default=None
    )
    """The approximate amount of members in this invite's guild.

    This is only present when `with_counts` is passed as `True` to the GET
    Invites endpoint.
    """


def _max_age_deserializer(age: int) -> datetime.timedelta:
    return datetime.timedelta(seconds=age) if age > 0 else None


@marshaller.marshallable()
@attr.s(slots=True, kw_only=True)
class InviteWithMetadata(Invite):
    """Extends the base `Invite` object with metadata.

    The metadata is only returned when getting an invite with
    guild permissions, rather than it's code.
    """

    uses: int = marshaller.attrib(deserializer=int, repr=True)
    """The amount of times this invite has been used."""

    max_uses: int = marshaller.attrib(deserializer=int, repr=True)
    """The limit for how many times this invite can be used before it expires.

    If set to `0` then this is unlimited.
    """

    max_age: typing.Optional[datetime.timedelta] = marshaller.attrib(deserializer=_max_age_deserializer)
    """The timedelta of how long this invite will be valid for.

    If set to `None` then this is unlimited.
    """

    is_temporary: bool = marshaller.attrib(raw_name="temporary", deserializer=bool, repr=True)
    """Whether this invite grants temporary membership."""

    created_at: datetime.datetime = marshaller.attrib(deserializer=conversions.parse_iso_8601_ts)
    """When this invite was created."""

    @property
    def expires_at(self) -> typing.Optional[datetime.datetime]:
        """When this invite should expire, if `InviteWithMetadata.max_age` is set.

        If this invite doesn't have a set expiry then this will be `None`.
        """
        if self.max_age:
            return self.created_at + self.max_age
        return None
